// Generated by CoffeeScript 1.3.3
(function() {
  var AFFILIATIONS, ALLOWED_ACCESS_MODELS, ALLOWED_PUBLISH_MODELS, AuthorizeSubscriber, BrowseInfo, BrowseNodeInfo, BrowseNodeItems, BrowseNodes, BrowseTopFollowedNodes, BrowseTopPublishedNodes, CreateNode, Element, ManageNodeAffiliations, ManageNodeConfiguration, ManageNodeSubscriptions, ModelOperation, ModeratorNotify, NODE_OWNER_TYPE_REGEXP, NS, NewModeratorNotify, Notify, OPERATIONS, Operation, PrivilegedOperation, Publish, PushInbox, Register, RemoveUser, ReplayArchive, RetractItems, RetrieveItems, RetrieveNodeAffiliations, RetrieveNodeConfiguration, RetrieveNodeSubscriptions, RetrieveUserAffiliations, RetrieveUserSubscriptions, Subscribe, Unsubscribe, async, canModerate, defaultConfiguration, defaultTopicConfiguration, errors, generateSubscriptionsNotifications, getNodeType, getNodeUser, groupByNode, inspect, isAffiliationAtLeast, logger, normalizeItem, runTransaction, uuid, _ref,
    _this = this,
    __hasProp = {}.hasOwnProperty,
    __extends = function(child, parent) { for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; } function ctor() { this.constructor = child; } ctor.prototype = parent.prototype; child.prototype = new ctor(); child.__super__ = parent.prototype; return child; },
    __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };

  logger = require('../logger').makeLogger('local/operations');

  inspect = require('util').inspect;

  _ref = require('../util'), getNodeUser = _ref.getNodeUser, getNodeType = _ref.getNodeType;

  async = require('async');

  uuid = require('node-uuid');

  errors = require('../errors');

  NS = require('../xmpp/ns');

  normalizeItem = require('../normalize').normalizeItem;

  Element = require('node-xmpp').Element;

  runTransaction = null;

  exports.setModel = function(model) {
    return runTransaction = model.transaction;
  };

  exports.checkCreateNode = null;

  defaultConfiguration = function(user) {
    return {
      posts: {
        title: "" + user + " Channel Posts",
        description: "A buddycloud channel",
        channelType: "personal",
        accessModel: "open",
        publishModel: "publishers",
        defaultAffiliation: "publisher"
      },
      status: {
        title: "" + user + " Status Updates",
        description: "M000D",
        accessModel: "open",
        publishModel: "publishers",
        defaultAffiliation: "member"
      },
      'geo/previous': {
        title: "" + user + " Previous Location",
        description: "Where " + user + " has been before",
        accessModel: "open",
        publishModel: "publishers",
        defaultAffiliation: "member"
      },
      'geo/current': {
        title: "" + user + " Current Location",
        description: "Where " + user + " is at now",
        accessModel: "open",
        publishModel: "publishers",
        defaultAffiliation: "member"
      },
      'geo/next': {
        title: "" + user + " Next Location",
        description: "Where " + user + " intends to go",
        accessModel: "open",
        publishModel: "publishers",
        defaultAffiliation: "member"
      },
      subscriptions: {
        title: "" + user + " Subscriptions",
        description: "Browse my interests",
        accessModel: "open",
        publishModel: "publishers",
        defaultAffiliation: "member"
      }
    };
  };

  defaultTopicConfiguration = function(user) {
    var _ref1;
    return {
      posts: {
        title: "" + user + " Topic Channel",
        description: "All about " + ((_ref1 = user.split('@')) != null ? _ref1[0] : void 0),
        channelType: "topic",
        accessModel: "open",
        publishModel: "subscribers",
        defaultAffiliation: "member"
      }
    };
  };

  NODE_OWNER_TYPE_REGEXP = /^\/user\/([^\/]+)\/?(.*)/;

  AFFILIATIONS = ['outcast', 'none', 'member', 'publisher', 'moderator', 'owner'];

  isAffiliationAtLeast = function(affiliation1, affiliation2) {
    var i1, i2;
    i1 = AFFILIATIONS.indexOf(affiliation1);
    i2 = AFFILIATIONS.indexOf(affiliation2);
    if (i2 < 0) {
      return false;
    } else {
      return i1 >= i2;
    }
  };

  canModerate = function(affiliation) {
    return affiliation === 'owner' || affiliation === 'moderator';
  };

  /*
  # Base Operations
  */


  Operation = (function() {

    function Operation(router, req) {
      var m;
      this.router = router;
      this.req = req;
      if ((this.req.node != null) && (m = this.req.node.match(NODE_OWNER_TYPE_REGEXP)) && m[2] === 'subscriptions') {
        this.subscriptionsNodeOwner = m[1];
      }
    }

    Operation.prototype.run = function(cb) {
      return cb(new errorsFeature.NotImplemented("Operation defined but not yet implemented"));
    };

    return Operation;

  })();

  ModelOperation = (function(_super) {

    __extends(ModelOperation, _super);

    function ModelOperation() {
      return ModelOperation.__super__.constructor.apply(this, arguments);
    }

    ModelOperation.prototype.run = function(cb) {
      var _this = this;
      return runTransaction(function(err, t) {
        var opName, _ref1;
        if (err) {
          return cb(err);
        }
        opName = _this.req.operation || ((_ref1 = _this.constructor) != null ? _ref1.name : void 0) || "?";
        return _this.transaction(t, function(err, results) {
          if (err) {
            logger.warn("Transaction " + opName + " rollback: " + err);
            return t.rollback(function() {
              return cb(err);
            });
          } else {
            return t.commit(function() {
              logger.debug("Transaction " + opName + " committed");
              return cb(null, results);
            });
          }
        });
      });
    };

    ModelOperation.prototype.transaction = function(t, cb) {
      return cb(null);
    };

    return ModelOperation;

  })(Operation);

  PrivilegedOperation = (function(_super) {

    __extends(PrivilegedOperation, _super);

    function PrivilegedOperation() {
      this.filterSubscription = __bind(this.filterSubscription, this);
      return PrivilegedOperation.__super__.constructor.apply(this, arguments);
    }

    PrivilegedOperation.prototype.transaction = function(t, cb) {
      var _this = this;
      return async.waterfall([
        function(cb2) {
          return _this.fetchActorAffiliation(t, cb2);
        }, function(cb2) {
          return _this.fetchNodeConfig(t, cb2);
        }, function(cb2) {
          return _this.checkAccessModel(t, cb2);
        }, function(cb2) {
          return _this.checkRequiredAffiliation(t, cb2);
        }
      ], function(err) {
        if (err) {
          return cb(err);
        }
        return _this.privilegedTransaction(t, cb);
      });
    };

    PrivilegedOperation.prototype.fetchActorAffiliation = function(t, cb) {
      var _this = this;
      if (!this.req.node) {
        return cb();
      }
      if (this.req.actor.indexOf('@') >= 0) {
        t.getAffiliation(this.req.node, this.req.actor, function(err, affiliation) {
          if (err) {
            return cb(err);
          }
          return _this.actorAffiliation = affiliation || 'none';
        });
      } else {
        t.getListenerAffiliations(this.req.node, this.req.actor, function(err, affiliations) {
          var affiliation, _i, _len, _results;
          if (err) {
            return cb(err);
          }
          _this.actorAffiliation = 'none';
          _results = [];
          for (_i = 0, _len = affiliations.length; _i < _len; _i++) {
            affiliation = affiliations[_i];
            if (affiliation !== _this.actorAffiliation && isAffiliationAtLeast(affiliation, _this.actorAffiliation)) {
              _results.push(_this.actorAffiliation = affiliation);
            } else {
              _results.push(void 0);
            }
          }
          return _results;
        });
      }
      if (canModerate(this.actorAffiliation)) {
        this.filterSubscription = this.filterSubscriptionModerator;
        this.filterAffiliation = this.filterAffiliationModerator;
      }
      return cb();
    };

    PrivilegedOperation.prototype.fetchNodeConfig = function(t, cb) {
      var _this = this;
      if (!this.req.node) {
        return cb();
      }
      return t.getConfig(this.req.node, function(err, config) {
        if (err) {
          return cb(err);
        }
        _this.nodeConfig = config;
        return cb();
      });
    };

    PrivilegedOperation.prototype.checkAccessModel = function(t, cb) {
      if (this.actorAffiliation === 'outcast') {
        return cb(new errors.Forbidden("Outcast"));
      }
      if (!this.requiredAffiliation) {
        if (this.nodeConfig.accessModel === 'open') {
          this.requiredAffiliation = 'none';
        } else {
          this.requiredAffiliation = 'member';
        }
      }
      return cb();
    };

    PrivilegedOperation.prototype.checkRequiredAffiliation = function(t, cb) {
      var requiredAffiliation, _ref1;
      if (((_ref1 = this.requiredAffiliation) != null ? _ref1.constructor : void 0) === Function) {
        requiredAffiliation = this.requiredAffiliation();
      } else {
        requiredAffiliation = this.requiredAffiliation;
      }
      if (!(requiredAffiliation != null) || isAffiliationAtLeast(this.actorAffiliation, requiredAffiliation)) {
        return cb();
      } else {
        return cb(new errors.Forbidden("Requires affiliation " + requiredAffiliation + " (you are " + this.actorAffiliation + ")"));
      }
    };

    PrivilegedOperation.prototype.checkPublishModel = function(t, cb) {
      var pass;
      pass = false;
      switch (this.nodeConfig.publishModel) {
        case 'open':
          pass = true;
          break;
        case 'members':
          pass = isAffiliationAtLeast(this.actorAffiliation, 'member');
          break;
        case 'publishers':
          pass = isAffiliationAtLeast(this.actorAffiliation, 'publisher');
          break;
        default:
          pass = this.actorAffiliation === 'owner';
      }
      if (pass) {
        return cb();
      } else if (this.nodeConfig.publishModel === 'subscribers') {
        return t.getSubscription(this.req.node, this.req.actor, function(err, subscription) {
          if (!err && subscription === 'subscribed') {
            return cb();
          } else {
            return cb(err || new errors.Forbidden("Only subscribers may publish"));
          }
        });
      } else {
        return cb(new errors.Forbidden("Only " + this.nodeConfig.publishModel + " may publish"));
      }
    };

    PrivilegedOperation.prototype.filterSubscription = function(subscription) {
      return subscription.jid === this.actor || subscription.subscription === 'subscribed';
    };

    PrivilegedOperation.prototype.filterAffiliation = function(affiliation) {
      return affiliation.affiliation !== 'outcast';
    };

    PrivilegedOperation.prototype.filterSubscriptionModerator = function(subscription) {
      return true;
    };

    PrivilegedOperation.prototype.filterAffiliationModerator = function(affiliation) {
      return true;
    };

    return PrivilegedOperation;

  })(ModelOperation);

  /*
  # Discrete Operations
  */


  BrowseInfo = (function(_super) {

    __extends(BrowseInfo, _super);

    function BrowseInfo() {
      return BrowseInfo.__super__.constructor.apply(this, arguments);
    }

    BrowseInfo.prototype.run = function(cb) {
      return cb(null, {
        features: [NS.DISCO_INFO, NS.DISCO_ITEMS, NS.REGISTER, NS.PUBSUB, NS.PUBSUB_OWNER],
        identities: [
          {
            category: "pubsub",
            type: "service",
            name: "XEP-0060 service"
          }, {
            category: "pubsub",
            type: "channels",
            name: "Channels service"
          }, {
            category: "pubsub",
            type: "inbox",
            name: "Channels inbox service"
          }
        ]
      });
    };

    return BrowseInfo;

  })(Operation);

  BrowseNodeInfo = (function(_super) {

    __extends(BrowseNodeInfo, _super);

    function BrowseNodeInfo() {
      return BrowseNodeInfo.__super__.constructor.apply(this, arguments);
    }

    BrowseNodeInfo.prototype.transaction = function(t, cb) {
      var _this = this;
      return t.getConfig(this.req.node, function(err, config) {
        return cb(err, {
          node: _this.req.node,
          features: [NS.DISCO_INFO, NS.DISCO_ITEMS, NS.REGISTER, NS.PUBSUB, NS.PUBSUB_OWNER],
          identities: [
            {
              category: "pubsub",
              type: "leaf",
              name: "XEP-0060 node"
            }, {
              category: "pubsub",
              type: "channel",
              name: "buddycloud channel"
            }
          ],
          config: config
        });
      });
    };

    return BrowseNodeInfo;

  })(PrivilegedOperation);

  BrowseNodes = (function(_super) {

    __extends(BrowseNodes, _super);

    function BrowseNodes() {
      return BrowseNodes.__super__.constructor.apply(this, arguments);
    }

    BrowseNodes.prototype.transaction = function(t, cb) {
      var rsm,
        _this = this;
      rsm = this.req.rsm;
      return this.fetchNodes(t, function(err, results) {
        if (err) {
          return cb(err);
        }
        results = rsm.cropResults(results, 'node');
        results.forEach(function(item) {
          return item.jid = _this.req.me;
        });
        return cb(null, results);
      });
    };

    BrowseNodes.prototype.fetchNodes = function(t, cb) {
      return t.listNodes(cb);
    };

    return BrowseNodes;

  })(ModelOperation);

  BrowseTopFollowedNodes = (function(_super) {

    __extends(BrowseTopFollowedNodes, _super);

    function BrowseTopFollowedNodes() {
      return BrowseTopFollowedNodes.__super__.constructor.apply(this, arguments);
    }

    BrowseTopFollowedNodes.prototype.fetchNodes = function(t, cb) {
      var max;
      max = this.req.rsm.max || 10;
      return t.getTopFollowedNodes(max, null, null, cb);
    };

    return BrowseTopFollowedNodes;

  })(BrowseNodes);

  BrowseTopPublishedNodes = (function(_super) {

    __extends(BrowseTopPublishedNodes, _super);

    function BrowseTopPublishedNodes() {
      return BrowseTopPublishedNodes.__super__.constructor.apply(this, arguments);
    }

    BrowseTopPublishedNodes.prototype.fetchNodes = function(t, cb) {
      var max;
      max = this.req.rsm.max || 10;
      return t.getTopPublishedNodes(max, null, null, cb);
    };

    return BrowseTopPublishedNodes;

  })(BrowseNodes);

  BrowseNodeItems = (function(_super) {

    __extends(BrowseNodeItems, _super);

    function BrowseNodeItems() {
      return BrowseNodeItems.__super__.constructor.apply(this, arguments);
    }

    BrowseNodeItems.prototype.privilegedTransaction = function(t, cb) {
      var _this = this;
      if (this.subscriptionsNodeOwner != null) {
        return t.getSubscriptions(this.subscriptionsNodeOwner, function(err, subscriptions) {
          var followee, m, present, results, subscription, subscriptionsByFollowee, _i, _len;
          if (err) {
            return cb(err);
          }
          subscriptionsByFollowee = {};
          for (_i = 0, _len = subscriptions.length; _i < _len; _i++) {
            subscription = subscriptions[_i];
            if ((m = subscription.node.match(NODE_OWNER_TYPE_REGEXP))) {
              followee = m[1];
              subscriptionsByFollowee[followee] = true;
            }
          }
          results = [];
          for (followee in subscriptionsByFollowee) {
            if (!__hasProp.call(subscriptionsByFollowee, followee)) continue;
            present = subscriptionsByFollowee[followee];
            results.push({
              name: followee,
              jid: _this.req.me,
              node: _this.req.node
            });
          }
          results.sort(function(result1, result2) {
            if (result1.name < result2.name) {
              return -1;
            } else if (result1.name > result2.name) {
              return 1;
            } else {
              return 0;
            }
          });
          results = _this.req.rsm.cropResults(results, 'name');
          return cb(null, results);
        });
      } else {
        return t.getItemIds(this.req.node, function(err, ids) {
          var results;
          if (err) {
            return cb(err);
          }
          ids = _this.req.rsm.cropResults(ids);
          results = ids.map(function(id) {
            return {
              name: id,
              jid: _this.req.me,
              node: _this.req.node
            };
          });
          results.node = _this.req.node;
          results.rsm = _this.req.rsm;
          return cb(null, results);
        });
      }
    };

    return BrowseNodeItems;

  })(PrivilegedOperation);

  Register = (function(_super) {

    __extends(Register, _super);

    function Register() {
      return Register.__super__.constructor.apply(this, arguments);
    }

    Register.prototype.run = function(cb) {
      var _this = this;
      return this.router.authorizeFor(this.req.me, this.req.actor, function(err, valid) {
        if (err) {
          return cb(err);
        }
        if (valid) {
          return ModelOperation.prototype.run.call(_this, cb);
        } else {
          return cb(new errors.NotAllowed("This is not the authoritative buddycloud-server for your domain"));
        }
      });
    };

    Register.prototype.transaction = function(t, cb) {
      var config, jobs, nodeType, user, _fn, _ref1,
        _this = this;
      user = this.req.actor;
      jobs = [];
      _ref1 = defaultConfiguration(user);
      _fn = function(nodeType, config) {
        return jobs.push(function(cb2) {
          var node;
          node = "/user/" + user + "/" + nodeType;
          config.creationDate = new Date().toISOString();
          return _this.createNodeWithConfig(t, node, config, cb2);
        });
      };
      for (nodeType in _ref1) {
        if (!__hasProp.call(_ref1, nodeType)) continue;
        config = _ref1[nodeType];
        _fn(nodeType, config);
      }
      return async.series(jobs, function(err) {
        return cb(err);
      });
    };

    Register.prototype.createNodeWithConfig = function(t, node, config, cb) {
      var created, user,
        _this = this;
      user = this.req.actor;
      created = false;
      return async.waterfall([
        function(cb2) {
          logger.info("creating " + node);
          return t.createNode(node, cb2);
        }, function(created_, cb2) {
          created = created_;
          return t.setAffiliation(node, user, 'owner', cb2);
        }, function(cb2) {
          return t.setSubscription(node, user, _this.req.sender, 'subscribed', cb2);
        }, function(cb2) {
          if (created) {
            return t.setConfig(node, config, cb2);
          } else {
            return cb2(null);
          }
        }
      ], cb);
    };

    return Register;

  })(ModelOperation);

  CreateNode = (function(_super) {

    __extends(CreateNode, _super);

    function CreateNode() {
      return CreateNode.__super__.constructor.apply(this, arguments);
    }

    CreateNode.prototype.transaction = function(t, cb) {
      var isTopic, nodePrefix, nodeUser, opts,
        _this = this;
      nodeUser = getNodeUser(this.req.node);
      if (!nodeUser) {
        return cb(new errors.BadRequest("Malformed node"));
      }
      if (nodeUser.split("@")[0].length < 1) {
        return cb(new errors.BadRequest("Malformed node name"));
      }
      isTopic = nodeUser !== this.req.actor;
      nodePrefix = "/user/" + nodeUser + "/";
      try {
        opts = {
          node: this.req.node,
          nodeUser: nodeUser,
          actor: this.req.actor
        };
        if (!(typeof exports.checkCreateNode === "function" ? exports.checkCreateNode(opts) : void 0)) {
          return cb(new errors.NotAllowed("Node creation not allowed"));
        }
      } catch (e) {
        return cb(e);
      }
      return async.waterfall([
        function(cb2) {
          return t.getOwnersByNodePrefix(nodePrefix, cb2);
        }, function(owners, cb2) {
          if (owners.length < 1 || owners.indexOf(_this.req.actor) >= 0) {
            return t.createNode(_this.req.node, cb2);
          } else {
            return cb2(new errors.NotAllowed("Nodes with prefix " + nodePrefix + " are already owned by " + (owners.join(', '))));
          }
        }, function(created, cb2) {
          if (created) {
            return cb2(null);
          } else {
            return cb2(new errors.Conflict("Node " + _this.req.node + " already exists"));
          }
        }, function(cb2) {
          var config, defaults, key, value;
          config = _this.req.config || {};
          config.creationDate = new Date().toISOString();
          if (isTopic) {
            defaults = defaultTopicConfiguration(nodeUser);
          } else {
            defaults = defaultConfiguration(nodeUser);
          }
          defaults = defaults[getNodeType(_this.req.node)];
          if (defaults) {
            for (key in defaults) {
              if (!__hasProp.call(defaults, key)) continue;
              value = defaults[key];
              if (!config) {
                config = {};
              }
              if (config[key] == null) {
                config[key] = value;
              }
            }
          }
          return t.setConfig(_this.req.node, config, cb2);
        }, function(cb2) {
          return t.setSubscription(_this.req.node, _this.req.actor, _this.req.sender, 'subscribed', cb2);
        }, function(cb2) {
          return t.setAffiliation(_this.req.node, _this.req.actor, 'owner', cb2);
        }
      ], cb);
    };

    return CreateNode;

  })(ModelOperation);

  Publish = (function(_super) {

    __extends(Publish, _super);

    function Publish() {
      return Publish.__super__.constructor.apply(this, arguments);
    }

    Publish.prototype.privilegedTransaction = function(t, cb) {
      var _this = this;
      if (this.subscriptionsNode != null) {
        return cb(new errors.NotAllowed("The subscriptions node is automagically populated"));
      }
      return async.waterfall([
        function(cb2) {
          return _this.checkPublishModel(t, cb2);
        }, function(cb2) {
          return async.series(_this.req.items.map(function(item) {
            return function(cb3) {
              return async.waterfall([
                function(cb4) {
                  if (item.id == null) {
                    item.id = uuid();
                    return cb4(null, null);
                  } else {
                    return t.getItem(_this.req.node, item.id, function(err, item) {
                      if (err && err.constructor === errors.NotFound) {
                        return cb4(null, null);
                      } else {
                        return cb4(err, item);
                      }
                    });
                  }
                }, function(oldItem, cb4) {
                  return normalizeItem(_this.req, oldItem, item, cb4);
                }, function(newItem, cb4) {
                  return t.writeItem(_this.req.node, newItem.id, newItem.el, function(err) {
                    return cb4(err, newItem.id);
                  });
                }
              ], cb3);
            };
          }), cb2);
        }
      ], cb);
    };

    Publish.prototype.notification = function() {
      return [
        {
          type: 'items',
          node: this.req.node,
          items: this.req.items
        }
      ];
    };

    return Publish;

  })(PrivilegedOperation);

  Subscribe = (function(_super) {

    __extends(Subscribe, _super);

    function Subscribe() {
      return Subscribe.__super__.constructor.apply(this, arguments);
    }

    Subscribe.prototype.transaction = function(t, cb) {
      var _this = this;
      return async.waterfall([
        function(cb2) {
          return _this.fetchActorAffiliation(t, cb2);
        }, function(cb2) {
          return _this.fetchNodeConfig(t, cb2);
        }, function(cb2) {
          var defaultAffiliation;
          if (_this.nodeConfig.accessModel === 'authorize') {
            _this.subscription = 'pending';
            return cb2();
          }
          _this.subscription = 'subscribed';
          defaultAffiliation = _this.nodeConfig.defaultAffiliation || 'none';
          if (!isAffiliationAtLeast(_this.actorAffiliation, defaultAffiliation)) {
            _this.affiliation = _this.nodeConfig.defaultAffiliation || 'member';
          }
          return _this.checkAccessModel(t, cb2);
        }
      ], function(err) {
        if (err) {
          return cb(err);
        }
        return _this.privilegedTransaction(t, cb);
      });
    };

    Subscribe.prototype.privilegedTransaction = function(t, cb) {
      var _this = this;
      return async.waterfall([
        function(cb2) {
          return t.setSubscription(_this.req.node, _this.req.actor, _this.req.sender, _this.subscription, cb2);
        }, function(cb2) {
          if (_this.affiliation) {
            return t.setAffiliation(_this.req.node, _this.req.actor, _this.affiliation, cb2);
          } else {
            return cb2();
          }
        }
      ], function(err) {
        return cb(err, {
          user: _this.req.actor,
          subscription: _this.subscription
        });
      });
    };

    Subscribe.prototype.notification = function() {
      var ns;
      ns = [
        {
          type: 'subscription',
          node: this.req.node,
          user: this.req.actor,
          subscription: this.subscription
        }
      ];
      if (this.affiliation) {
        ns.push({
          type: 'affiliation',
          node: this.req.node,
          user: this.req.actor,
          affiliation: this.affiliation
        });
      }
      return ns;
    };

    Subscribe.prototype.moderatorNotification = function() {
      if (this.subscription === 'pending') {
        return {
          type: 'authorizationPrompt',
          node: this.req.node,
          user: this.req.actor
        };
      }
    };

    return Subscribe;

  })(PrivilegedOperation);

  Unsubscribe = (function(_super) {

    __extends(Unsubscribe, _super);

    function Unsubscribe() {
      return Unsubscribe.__super__.constructor.apply(this, arguments);
    }

    Unsubscribe.prototype.transaction = function(t, cb) {
      var _this = this;
      if (this.req.node.indexOf("/user/" + this.req.actor + "/") === 0) {
        return cb(new errors.Forbidden("You may not unsubscribe from your own nodes"));
      }
      return async.waterfall([
        function(cb2) {
          return t.setSubscription(_this.req.node, _this.req.actor, _this.req.sender, 'none', cb2);
        }, function(cb2) {
          return _this.fetchActorAffiliation(t, cb2);
        }, function(cb2) {
          return _this.fetchNodeConfig(t, cb2);
        }, function(cb2) {
          if (isAffiliationAtLeast(_this.nodeConfig.defaultAffiliation, _this.actorAffiliation) && _this.actorAffiliation !== 'outcast') {
            _this.actorAffiliation = 'none';
            return t.setAffiliation(_this.req.node, _this.req.actor, 'none', cb2);
          } else {
            return cb2();
          }
        }
      ], cb);
    };

    Unsubscribe.prototype.notification = function() {
      return [
        {
          type: 'subscription',
          node: this.req.node,
          user: this.req.actor,
          subscription: 'none'
        }, {
          type: 'affiliation',
          node: this.req.node,
          user: this.req.actor,
          affiliation: this.actorAffiliation
        }
      ];
    };

    return Unsubscribe;

  })(PrivilegedOperation);

  RetrieveItems = (function(_super) {

    __extends(RetrieveItems, _super);

    function RetrieveItems() {
      return RetrieveItems.__super__.constructor.apply(this, arguments);
    }

    RetrieveItems.prototype.run = function(cb) {
      var _this = this;
      if (this.subscriptionsNodeOwner != null) {
        return this.router.authorizeFor(this.req.me, this.subscriptionsNodeOwner, function(err, valid) {
          if (err) {
            return cb(err);
          }
          if (valid) {
            _this.privilegedTransaction = _this.retrieveSubscriptionsItems;
          }
          return PrivilegedOperation.prototype.run.call(_this, cb);
        });
      } else {
        return RetrieveItems.__super__.run.apply(this, arguments);
      }
    };

    RetrieveItems.prototype.privilegedTransaction = function(t, cb) {
      var fetchItemIds, node, rsm,
        _this = this;
      node = this.req.node;
      rsm = this.req.rsm;
      if (this.req.itemIds != null) {
        fetchItemIds = function(cb2) {
          return cb2(null, _this.req.itemIds);
        };
      } else {
        fetchItemIds = function(cb2) {
          return t.getItemIds(node, cb2);
        };
      }
      return fetchItemIds(function(err, ids) {
        ids = rsm.cropResults(ids);
        return async.series(ids.map(function(id) {
          return function(cb2) {
            return t.getItem(node, id, function(err, el) {
              if (err) {
                return cb2(err);
              }
              return cb2(null, {
                id: id,
                el: el
              });
            });
          };
        }), function(err, results) {
          if (err) {
            return cb(err);
          } else {
            results.node = node;
            results.rsm = rsm;
            return cb(null, results);
          }
        });
      });
    };

    RetrieveItems.prototype.retrieveSubscriptionsItems = function(t, cb) {
      var _this = this;
      return async.waterfall([
        function(cb2) {
          return t.getSubscriptions(_this.subscriptionsNodeOwner, cb2);
        }, function(subscriptions, cb2) {
          var followee, followeeSubscriptions, m, results, subscription, subscriptionsByFollowee, _i, _len;
          subscriptions = subscriptions.filter(function(subscription) {
            return subscription.subscription === 'subscribed';
          });
          subscriptionsByFollowee = {};
          for (_i = 0, _len = subscriptions.length; _i < _len; _i++) {
            subscription = subscriptions[_i];
            if ((m = subscription.node.match(NODE_OWNER_TYPE_REGEXP))) {
              followee = m[1];
              if (subscriptionsByFollowee[followee] == null) {
                subscriptionsByFollowee[followee] = [];
              }
              subscriptionsByFollowee[followee].push(subscription);
            }
          }
          results = [];
          for (followee in subscriptionsByFollowee) {
            if (!__hasProp.call(subscriptionsByFollowee, followee)) continue;
            followeeSubscriptions = subscriptionsByFollowee[followee];
            results.push({
              id: followee,
              subscriptions: followeeSubscriptions
            });
          }
          results.sort(function(result1, result2) {
            if (result1.id < result2.id) {
              return -1;
            } else if (result1.id > result2.id) {
              return 1;
            } else {
              return 0;
            }
          });
          results = _this.req.rsm.cropResults(results, 'id');
          return async.forEachSeries(results, function(result, cb3) {
            return async.forEach(result.subscriptions, function(subscription, cb4) {
              return t.getAffiliation(subscription.node, _this.subscriptionsNodeOwner, function(err, affiliation) {
                var _ref1;
                if ((_ref1 = subscription.affiliation) == null) {
                  subscription.affiliation = affiliation;
                }
                return cb4(err);
              });
            }, cb3);
          }, function(err) {
            return cb2(err, results);
          });
        }, function(results, cb2) {
          var item, itemAttrs, subscription, _i, _j, _len, _len1, _ref1, _ref2, _ref3;
          for (_i = 0, _len = results.length; _i < _len; _i++) {
            item = results[_i];
            item.el = new Element('query', {
              xmlns: NS.DISCO_ITEMS,
              'xmlns:pubsub': NS.PUBSUB
            });
            _ref1 = item.subscriptions;
            for (_j = 0, _len1 = _ref1.length; _j < _len1; _j++) {
              subscription = _ref1[_j];
              itemAttrs = {
                jid: _this.subscriptionsNodeOwner,
                node: subscription.node
              };
              if ((_ref2 = itemAttrs['pubsub:subscription']) == null) {
                itemAttrs['pubsub:subscription'] = subscription.subscription;
              }
              if ((_ref3 = itemAttrs['pubsub:affiliation']) == null) {
                itemAttrs['pubsub:affiliation'] = subscription.affiliation;
              }
              item.el.c('item', itemAttrs);
            }
            delete item.subscriptions;
          }
          results.rsm = _this.req.rsm;
          results.node = _this.req.node;
          return cb2(null, results);
        }
      ], cb);
    };

    return RetrieveItems;

  })(PrivilegedOperation);

  RetractItems = (function(_super) {

    __extends(RetractItems, _super);

    function RetractItems() {
      return RetractItems.__super__.constructor.apply(this, arguments);
    }

    RetractItems.prototype.privilegedTransaction = function(t, cb) {
      var _this = this;
      return async.waterfall([
        function(cb2) {
          if (isAffiliationAtLeast(_this.actorAffiliation, 'moderator')) {
            return cb2();
          } else {
            return _this.checkItemsAuthor(t, cb2);
          }
        }, function(cb2) {
          return async.forEach(_this.req.items, function(id, cb3) {
            return t.deleteItem(_this.req.node, id, cb3);
          }, cb2);
        }
      ], cb);
    };

    RetractItems.prototype.checkItemsAuthor = function(t, cb) {
      var _this = this;
      return async.forEachSeries(this.req.items, function(id, cb2) {
        return t.getItem(_this.req.node, id, function(err, el) {
          var author, _ref1, _ref2;
          if ((err != null ? err.constructor : void 0) === errors.NotFound) {
            return cb2();
          } else if (err) {
            return cb2(err);
          }
          author = (el != null ? el.is('entry') : void 0) && ((_ref1 = el.getChild('author')) != null ? (_ref2 = _ref1.getChild('uri')) != null ? _ref2.getText() : void 0 : void 0);
          if (author === ("acct:" + _this.req.actor)) {
            return cb2();
          } else {
            return cb2(new errors.NotAllowed("You may not retract other people's posts"));
          }
        });
      }, cb);
    };

    return RetractItems;

  })(PrivilegedOperation);

  RetrieveUserSubscriptions = (function(_super) {

    __extends(RetrieveUserSubscriptions, _super);

    function RetrieveUserSubscriptions() {
      return RetrieveUserSubscriptions.__super__.constructor.apply(this, arguments);
    }

    RetrieveUserSubscriptions.prototype.transaction = function(t, cb) {
      var rsm;
      rsm = this.req.rsm;
      return t.getSubscriptions(this.req.actor, function(err, subscriptions) {
        if (err) {
          return cb(err);
        }
        subscriptions = rsm.cropResults(subscriptions, 'node');
        return cb(null, subscriptions);
      });
    };

    return RetrieveUserSubscriptions;

  })(ModelOperation);

  RetrieveUserAffiliations = (function(_super) {

    __extends(RetrieveUserAffiliations, _super);

    function RetrieveUserAffiliations() {
      return RetrieveUserAffiliations.__super__.constructor.apply(this, arguments);
    }

    RetrieveUserAffiliations.prototype.transaction = function(t, cb) {
      var rsm;
      rsm = this.req.rsm;
      return t.getAffiliations(this.req.actor, function(err, affiliations) {
        if (err) {
          return cb(err);
        }
        affiliations = rsm.cropResults(affiliations, 'node');
        return cb(null, affiliations);
      });
    };

    return RetrieveUserAffiliations;

  })(ModelOperation);

  RetrieveNodeSubscriptions = (function(_super) {

    __extends(RetrieveNodeSubscriptions, _super);

    function RetrieveNodeSubscriptions() {
      return RetrieveNodeSubscriptions.__super__.constructor.apply(this, arguments);
    }

    RetrieveNodeSubscriptions.prototype.privilegedTransaction = function(t, cb) {
      var rsm,
        _this = this;
      rsm = this.req.rsm;
      return t.getSubscribers(this.req.node, function(err, subscriptions) {
        if (err) {
          return cb(err);
        }
        subscriptions = subscriptions.filter(_this.filterSubscription);
        subscriptions = rsm.cropResults(subscriptions, 'user');
        return cb(null, subscriptions);
      });
    };

    return RetrieveNodeSubscriptions;

  })(PrivilegedOperation);

  RetrieveNodeAffiliations = (function(_super) {

    __extends(RetrieveNodeAffiliations, _super);

    function RetrieveNodeAffiliations() {
      return RetrieveNodeAffiliations.__super__.constructor.apply(this, arguments);
    }

    RetrieveNodeAffiliations.prototype.privilegedTransaction = function(t, cb) {
      var rsm,
        _this = this;
      rsm = this.req.rsm;
      return t.getAffiliated(this.req.node, function(err, affiliations) {
        if (err) {
          return cb(err);
        }
        affiliations = affiliations.filter(_this.filterAffiliation);
        affiliations = rsm.cropResults(affiliations, 'user');
        return cb(null, affiliations);
      });
    };

    return RetrieveNodeAffiliations;

  })(PrivilegedOperation);

  RetrieveNodeConfiguration = (function(_super) {

    __extends(RetrieveNodeConfiguration, _super);

    function RetrieveNodeConfiguration() {
      return RetrieveNodeConfiguration.__super__.constructor.apply(this, arguments);
    }

    RetrieveNodeConfiguration.prototype.transaction = function(t, cb) {
      return t.getConfig(this.req.node, function(err, config) {
        return cb(err, {
          config: config
        });
      });
    };

    return RetrieveNodeConfiguration;

  })(PrivilegedOperation);

  ManageNodeSubscriptions = (function(_super) {

    __extends(ManageNodeSubscriptions, _super);

    function ManageNodeSubscriptions() {
      this.requiredAffiliation = __bind(this.requiredAffiliation, this);
      return ManageNodeSubscriptions.__super__.constructor.apply(this, arguments);
    }

    ManageNodeSubscriptions.prototype.requiredAffiliation = function() {
      if (this.nodeConfig.channelType === 'topic') {
        return 'moderator';
      } else {
        return 'owner';
      }
    };

    ManageNodeSubscriptions.prototype.privilegedTransaction = function(t, cb) {
      var defaultAffiliation,
        _this = this;
      defaultAffiliation = null;
      return async.waterfall([
        function(cb2) {
          return t.getConfig(_this.req.node, function(error, config) {
            defaultAffiliation = config.defaultAffiliation || 'member';
            return cb2(error);
          });
        }, function(cb2) {
          return async.forEach(_this.req.subscriptions, function(_arg, cb3) {
            var subscription, user;
            user = _arg.user, subscription = _arg.subscription;
            return async.waterfall([
              function(cb4) {
                if (_this.req.node.indexOf("/user/" + user + "/") === 0 && subscription !== 'subscribed') {
                  return cb4(new errors.Forbidden("You may not unsubscribe the owner"));
                } else {
                  return t.setSubscription(_this.req.node, user, null, subscription, cb4);
                }
              }, function(cb4) {
                return t.getAffiliation(_this.req.node, user, cb4);
              }, function(affiliation, cb4) {
                if (affiliation === 'none') {
                  return t.setAffiliation(_this.req.node, user, defaultAffiliation, cb4);
                } else {
                  return cb4();
                }
              }
            ], cb3);
          }, cb2);
        }
      ], cb);
    };

    ManageNodeSubscriptions.prototype.notification = function() {
      var _this = this;
      return this.req.subscriptions.map(function(_arg) {
        var subscription, user;
        user = _arg.user, subscription = _arg.subscription;
        return {
          type: 'subscription',
          node: _this.req.node,
          user: user,
          subscription: subscription
        };
      });
    };

    return ManageNodeSubscriptions;

  })(PrivilegedOperation);

  ManageNodeAffiliations = (function(_super) {

    __extends(ManageNodeAffiliations, _super);

    function ManageNodeAffiliations() {
      this.requiredAffiliation = __bind(this.requiredAffiliation, this);
      return ManageNodeAffiliations.__super__.constructor.apply(this, arguments);
    }

    ManageNodeAffiliations.prototype.requiredAffiliation = function() {
      if (this.nodeConfig.channelType === 'topic') {
        return 'moderator';
      } else {
        return 'owner';
      }
    };

    ManageNodeAffiliations.prototype.privilegedTransaction = function(t, cb) {
      var _this = this;
      this.newModerators = [];
      return async.series(this.req.affiliations.map(function(_arg) {
        var affiliation, user;
        user = _arg.user, affiliation = _arg.affiliation;
        return function(cb2) {
          return async.waterfall([
            function(cb3) {
              return t.getAffiliation(_this.req.node, user, cb3);
            }, function(oldAffiliation, cb3) {
              if (oldAffiliation === affiliation) {
                return cb3();
              } else {
                if (oldAffiliation !== 'owner' && affiliation === 'owner' && _this.actorAffiliation !== 'owner') {
                  return cb3(new errors.Forbidden("You may not elect a new owner"));
                } else {
                  return async.series([
                    function(cb4) {
                      if (!canModerate(oldAffiliation) && canModerate(affiliation)) {
                        return t.getSubscriptionListener(_this.req.node, user, function(err, listener) {
                          if ((!err) && listener) {
                            _this.newModerators.push({
                              user: user,
                              listener: listener,
                              node: _this.req.node
                            });
                          }
                          return cb4(err);
                        });
                      } else {
                        return cb4();
                      }
                    }, function(cb4) {
                      if (_this.req.node.indexOf("/user/" + user + "/") === 0 && affiliation !== 'owner') {
                        return cb4(new errors.Forbidden("You may not demote the owner"));
                      } else {
                        return t.setAffiliation(_this.req.node, user, affiliation, cb4);
                      }
                    }
                  ], function(err) {
                    return cb3(err);
                  });
                }
              }
            }
          ], cb2);
        };
      }), cb);
    };

    ManageNodeAffiliations.prototype.notification = function() {
      var affiliations,
        _this = this;
      return affiliations = this.req.affiliations.map(function(_arg) {
        var affiliation, user;
        user = _arg.user, affiliation = _arg.affiliation;
        return {
          type: 'affiliation',
          node: _this.req.node,
          user: user,
          affiliation: affiliation
        };
      });
    };

    return ManageNodeAffiliations;

  })(PrivilegedOperation);

  ALLOWED_ACCESS_MODELS = ['open', 'whitelist', 'authorize'];

  ALLOWED_PUBLISH_MODELS = ['open', 'subscribers', 'publishers'];

  ManageNodeConfiguration = (function(_super) {

    __extends(ManageNodeConfiguration, _super);

    function ManageNodeConfiguration() {
      return ManageNodeConfiguration.__super__.constructor.apply(this, arguments);
    }

    ManageNodeConfiguration.prototype.requiredAffiliation = 'owner';

    ManageNodeConfiguration.prototype.run = function(cb) {
      if ((this.req.config.accessModel != null) && ALLOWED_ACCESS_MODELS.indexOf(this.req.config.accessModel) < 0) {
        return cb(new errors.BadRequest("Invalid access model"));
      } else if ((this.req.config.publishModel != null) && ALLOWED_PUBLISH_MODELS.indexOf(this.req.config.publishModel) < 0) {
        return cb(new errors.BadRequest("Invalid publish model"));
      } else if (this.req.config.creationDate != null) {
        return cb(new errors.BadRequest("Cannot set creation date"));
      } else {
        return ManageNodeConfiguration.__super__.run.call(this, cb);
      }
    };

    ManageNodeConfiguration.prototype.privilegedTransaction = function(t, cb) {
      return t.setConfig(this.req.node, this.req.config, cb);
    };

    ManageNodeConfiguration.prototype.notification = function() {
      return [
        {
          type: 'config',
          node: this.req.node,
          config: this.req.config
        }
      ];
    };

    return ManageNodeConfiguration;

  })(PrivilegedOperation);

  RemoveUser = (function(_super) {

    __extends(RemoveUser, _super);

    function RemoveUser() {
      return RemoveUser.__super__.constructor.apply(this, arguments);
    }

    RemoveUser.prototype.transaction = function(t, cb) {
      var _this = this;
      return t.getUserRemoteSubscriptions(this.req.actor, function(err, subscriptions) {
        if (err) {
          return cb(err);
        }
        return t.clearUserSubscriptions(_this.req.actor, function(err) {
          return cb(err, subscriptions);
        });
      });
    };

    return RemoveUser;

  })(ModelOperation);

  AuthorizeSubscriber = (function(_super) {

    __extends(AuthorizeSubscriber, _super);

    function AuthorizeSubscriber() {
      this.requiredAffiliation = __bind(this.requiredAffiliation, this);
      return AuthorizeSubscriber.__super__.constructor.apply(this, arguments);
    }

    AuthorizeSubscriber.prototype.requiredAffiliation = function() {
      if (this.nodeConfig.channelType === 'topic') {
        return 'moderator';
      } else {
        return 'owner';
      }
    };

    AuthorizeSubscriber.prototype.privilegedTransaction = function(t, cb) {
      var _this = this;
      if (this.req.allow) {
        this.subscription = 'subscribed';
        if (!isAffiliationAtLeast(this.actorAffiliation, this.nodeConfig.defaultAffiliation)) {
          this.affiliation = this.nodeConfig.defaultAffiliation || 'member';
        }
      } else {
        this.subscription = 'none';
      }
      return async.waterfall([
        function(cb2) {
          return t.setSubscription(_this.req.node, _this.req.user, _this.req.sender, _this.subscription, cb2);
        }, function(cb2) {
          if (_this.affiliation) {
            return t.setAffiliation(_this.req.node, _this.req.user, _this.affiliation, cb2);
          } else {
            return cb2();
          }
        }
      ], function(err) {
        return cb(err);
      });
    };

    AuthorizeSubscriber.prototype.notification = function() {
      var ns;
      ns = [
        {
          type: 'subscription',
          node: this.req.node,
          user: this.req.user,
          subscription: this.subscription
        }
      ];
      if (this.affiliation) {
        ns.push({
          type: 'affiliation',
          node: this.req.node,
          user: this.req.user,
          affiliation: this.affiliation
        });
      }
      return ns;
    };

    return AuthorizeSubscriber;

  })(PrivilegedOperation);

  ReplayArchive = (function(_super) {

    __extends(ReplayArchive, _super);

    function ReplayArchive() {
      return ReplayArchive.__super__.constructor.apply(this, arguments);
    }

    ReplayArchive.prototype.transaction = function(t, cb) {
      var max, sent, _ref1,
        _this = this;
      max = ((_ref1 = this.req.rsm) != null ? _ref1.max : void 0) || 50;
      sent = 0;
      return async.waterfall([
        function(cb2) {
          return t.walkListenerArchive(_this.req.sender, _this.req.start, _this.req.end, max, function(results) {
            if (sent < max) {
              results.sort(function(a, b) {
                if (a.updated < b.updated) {
                  return -1;
                } else if (a.updated > b.updated) {
                  return 1;
                } else {
                  return 0;
                }
              });
              results = results.slice(0, max - sent);
              _this.sendNotification(results);
              return sent += results.length;
            }
          }, cb2);
        }, function(cb2) {
          sent = 0;
          return t.walkModeratorAuthorizationRequests(_this.req.sender, function(req) {
            if (sent < max) {
              req.type = 'authorizationPrompt';
              _this.sendNotification(req);
              return sent++;
            }
          }, cb2);
        }
      ], cb);
    };

    ReplayArchive.prototype.sendNotification = function(results) {
      var notification;
      notification = Object.create(results);
      notification.listener = this.req.fullSender;
      notification.replay = true;
      return this.router.notify(notification);
    };

    return ReplayArchive;

  })(ModelOperation);

  PushInbox = (function(_super) {

    __extends(PushInbox, _super);

    function PushInbox() {
      return PushInbox.__super__.constructor.apply(this, arguments);
    }

    PushInbox.prototype.transaction = function(t, cb) {
      var newModerators, newNodes, notification, unsubscribedNodes,
        _this = this;
      notification = [];
      newNodes = [];
      newModerators = [];
      unsubscribedNodes = {};
      return async.waterfall([
        function(cb2) {
          logger.debug("pushUpdates: " + (inspect(_this.req)));
          return async.filter(_this.req, function(update, cb3) {
            if (update.type === 'subscription' && (update.listener != null)) {
              return t.createNode(update.node, function(err, created) {
                if (created) {
                  newNodes.push(update.node);
                }
                return cb3(!err);
              });
            } else {
              return t.nodeExists(update.node, function(err, exists) {
                return cb3(!err && exists);
              });
            }
          }, function(updates) {
            return cb2(null, updates);
          });
        }, function(updates, cb2) {
          logger.debug("pushFilteredUpdates: " + (inspect(updates)));
          return async.forEach(updates, function(update, cb3) {
            var affiliation, config, items, listener, node, subscription, user;
            switch (update.type) {
              case 'items':
                notification.push(update);
                node = update.node, items = update.items;
                return async.forEach(items, function(item, cb4) {
                  var el, id;
                  id = item.id, el = item.el;
                  return t.writeItem(node, id, el, cb4);
                }, cb3);
              case 'subscription':
                notification.push(update);
                node = update.node, user = update.user, listener = update.listener, subscription = update.subscription;
                if (subscription !== 'subscribed') {
                  unsubscribedNodes[node] = true;
                }
                return t.setSubscription(node, user, listener, subscription, cb3);
              case 'affiliation':
                notification.push(update);
                node = update.node, user = update.user, affiliation = update.affiliation;
                return t.getAffiliation(node, user, function(err, oldAffiliation) {
                  var _this = this;
                  if (err) {
                    return cb3(err);
                  }
                  return async.series([
                    function(cb4) {
                      if (!canModerate(oldAffiliation) && canModerate(affiliation)) {
                        return t.getSubscriptionListener(node, user, function(err, listener) {
                          if ((!err) && listener) {
                            newModerators.push({
                              user: user,
                              listener: listener,
                              node: node
                            });
                          }
                          return cb4(err);
                        });
                      } else {
                        return cb4();
                      }
                    }, function(cb4) {
                      return t.setAffiliation(node, user, affiliation, cb4);
                    }
                  ], function(err) {
                    return cb3(err);
                  });
                });
              case 'config':
                notification.push(update);
                node = update.node, config = update.config;
                return t.setConfig(node, config, cb3);
              default:
                return cb3(new errors.InternalServerError("Bogus push update type: " + update.type));
            }
          }, cb2);
        }, function(cb2) {
          _this.notification = notification;
          if (newNodes.length > 0) {
            _this.newNodes = newNodes;
          }
          if (newModerators.length > 0) {
            _this.newModerators = newModerators;
          }
          return cb2();
        }, function(cb2) {
          var checks, node, _;
          checks = [];
          for (node in unsubscribedNodes) {
            if (!__hasProp.call(unsubscribedNodes, node)) continue;
            _ = unsubscribedNodes[node];
            checks.push(function(cb3) {
              return t.getNodeListeners(node, function(err, listeners) {
                if (err) {
                  return cb3(err);
                }
                if (!((listeners != null) && listeners.length > 0)) {
                  return t.purgeNode(node, cb3);
                }
              });
            });
          }
          return async.parallel(checks, cb2);
        }
      ], cb);
    };

    return PushInbox;

  })(ModelOperation);

  Notify = (function(_super) {

    __extends(Notify, _super);

    function Notify() {
      return Notify.__super__.constructor.apply(this, arguments);
    }

    Notify.prototype.transaction = function(t, cb) {
      var _this = this;
      return async.waterfall([
        function(cb2) {
          return async.forEach(_this.req, function(update, cb3) {
            var m, subscriber;
            m = update.node.match(NODE_OWNER_TYPE_REGEXP);
            if (update.type === 'items' && (m != null ? m[2] : void 0) === 'subscriptions') {
              subscriber = m[1];
              return t.getSubscriptions(subscriber, function(err, subscriptions) {
                if (err) {
                  return cb3(err);
                }
                return async.map(update.items, function(_arg, cb4) {
                  var affiliations, el, id, userSubscriptions;
                  id = _arg.id, el = _arg.el;
                  if (el) {
                    return cb4(null, {
                      id: id,
                      el: el
                    });
                  }
                  userSubscriptions = subscriptions.filter(function(subscription) {
                    return subscription.node.indexOf("/user/" + id + "/") === 0;
                  });
                  affiliations = {};
                  return async.forEach(userSubscriptions, function(subscription, cb5) {
                    return t.getAffiliation(subscriber, subscription.node, function(err, affiliation) {
                      affiliations[subscription.node] = affiliation;
                      return cb5(err);
                    });
                  }, function(err) {
                    var itemAttrs, subscription, _i, _len, _ref1, _ref2;
                    el = new Element('query', {
                      xmlns: NS.DISCO_ITEMS,
                      'xmlns:pubsub': NS.PUBSUB
                    });
                    for (_i = 0, _len = userSubscriptions.length; _i < _len; _i++) {
                      subscription = userSubscriptions[_i];
                      itemAttrs = {
                        jid: subscriber,
                        node: subscription.node
                      };
                      if ((_ref1 = itemAttrs['pubsub:subscription']) == null) {
                        itemAttrs['pubsub:subscription'] = subscription.subscription;
                      }
                      if ((_ref2 = itemAttrs['pubsub:affiliation']) == null) {
                        itemAttrs['pubsub:affiliation'] = affiliations[subscription.node];
                      }
                      el.c('item', itemAttrs);
                    }
                    return cb4(null, {
                      id: id,
                      el: el
                    });
                  });
                }, function(err, items) {
                  if (err) {
                    return cb3(err);
                  }
                  update.items = items;
                  return cb3();
                });
              });
            } else {
              return cb3();
            }
          }, cb2);
        }, function(cb2) {
          logger.debug("notifyNotification: " + (inspect(_this.req)));
          return t.getNodeListeners(_this.req.node, cb2);
        }, function(listeners, cb2) {
          var update, _i, _len, _ref1;
          _ref1 = _this.req;
          for (_i = 0, _len = _ref1.length; _i < _len; _i++) {
            update = _ref1[_i];
            if (update.type === 'subscription' && listeners.indexOf(update.user) < 0) {
              listeners.push(update.user);
            }
          }
          return cb2(null, listeners);
        }, function(listeners, cb2) {
          var moderatorListeners, otherListeners;
          moderatorListeners = [];
          otherListeners = [];
          return async.forEach(listeners, function(listener, cb3) {
            return t.getAffiliation(_this.req.node, listener, function(err, affiliation) {
              if (err) {
                return cb3(err);
              }
              if (canModerate(affiliation)) {
                moderatorListeners.push(listener);
              } else {
                otherListeners.push(listener);
              }
              return cb3();
            });
          }, function(err) {
            return cb2(err, moderatorListeners, otherListeners);
          });
        }, function(moderatorListeners, otherListeners, cb2) {
          var listener, notification, req, _i, _j, _len, _len1;
          if (moderatorListeners.length > 0) {
            for (_i = 0, _len = moderatorListeners.length; _i < _len; _i++) {
              listener = moderatorListeners[_i];
              notification = Object.create(_this.req);
              notification.listener = listener;
              _this.router.notify(notification);
            }
          }
          if (otherListeners.length > 0) {
            req = _this.req.filter(function(update) {
              switch (update.type) {
                case 'subscription':
                  return PrivilegedOperation.prototype.filterSubscription(update);
                case 'affiliation':
                  return PrivilegedOperation.prototype.filterAffiliation(update);
                default:
                  return true;
              }
            });
            if (req.length > 0) {
              for (_j = 0, _len1 = otherListeners.length; _j < _len1; _j++) {
                listener = otherListeners[_j];
                notification = Object.create(req);
                notification.node = _this.req.node;
                notification.listener = listener;
                _this.router.notify(notification);
              }
            }
          }
          return cb2();
        }
      ], cb);
    };

    return Notify;

  })(ModelOperation);

  ModeratorNotify = (function(_super) {

    __extends(ModeratorNotify, _super);

    function ModeratorNotify() {
      return ModeratorNotify.__super__.constructor.apply(this, arguments);
    }

    ModeratorNotify.prototype.transaction = function(t, cb) {
      var _this = this;
      logger.debug("moderatorNotifyNotification: " + (inspect(this.req)));
      return t.getNodeModeratorListeners(this.req.node, function(err, listeners) {
        var listener, notification, _i, _len;
        if (err) {
          return cb(err);
        }
        for (_i = 0, _len = listeners.length; _i < _len; _i++) {
          listener = listeners[_i];
          notification = Object.create(_this.req);
          notification.listener = listener;
          _this.router.notify(notification);
        }
        return cb();
      });
    };

    return ModeratorNotify;

  })(ModelOperation);

  NewModeratorNotify = (function(_super) {

    __extends(NewModeratorNotify, _super);

    function NewModeratorNotify() {
      return NewModeratorNotify.__super__.constructor.apply(this, arguments);
    }

    NewModeratorNotify.prototype.privilegedTransaction = function(t, cb) {
      var _this = this;
      return async.parallel([
        function(cb2) {
          return t.getPending(_this.req.node, cb2);
        }, function(cb2) {
          return t.getOutcast(_this.req.node, cb2);
        }
      ], function(err, _arg) {
        var bannedUsers, notification, pendingUsers, user, _i, _j, _len, _len1;
        pendingUsers = _arg[0], bannedUsers = _arg[1];
        if (err) {
          return cb(err);
        }
        notification = [];
        notification.node = _this.req.node;
        notification.listener = _this.req.listener;
        for (_i = 0, _len = pendingUsers.length; _i < _len; _i++) {
          user = pendingUsers[_i];
          notification.push({
            type: 'subscription',
            node: _this.req.node,
            user: user,
            subscription: 'pending'
          });
        }
        for (_j = 0, _len1 = bannedUsers.length; _j < _len1; _j++) {
          user = bannedUsers[_j];
          notification.push({
            type: 'affiliation',
            node: _this.req.node,
            user: user,
            affiliation: 'outcast'
          });
        }
        _this.router.notify(notification);
        return cb();
      });
    };

    return NewModeratorNotify;

  })(PrivilegedOperation);

  OPERATIONS = {
    'browse-info': BrowseInfo,
    'browse-node-info': BrowseNodeInfo,
    'browse-nodes': BrowseNodes,
    'browse-top-followed-nodes': BrowseTopFollowedNodes,
    'browse-top-published-nodes': BrowseTopPublishedNodes,
    'browse-node-items': BrowseNodeItems,
    'register-user': Register,
    'create-node': CreateNode,
    'publish-node-items': Publish,
    'subscribe-node': Subscribe,
    'unsubscribe-node': Unsubscribe,
    'retrieve-node-items': RetrieveItems,
    'retract-node-items': RetractItems,
    'retrieve-user-subscriptions': RetrieveUserSubscriptions,
    'retrieve-user-affiliations': RetrieveUserAffiliations,
    'retrieve-node-subscriptions': RetrieveNodeSubscriptions,
    'retrieve-node-affiliations': RetrieveNodeAffiliations,
    'retrieve-node-configuration': RetrieveNodeConfiguration,
    'manage-node-subscriptions': ManageNodeSubscriptions,
    'manage-node-affiliations': ManageNodeAffiliations,
    'manage-node-configuration': ManageNodeConfiguration,
    'remove-user': RemoveUser,
    'confirm-subscriber-authorization': AuthorizeSubscriber,
    'replay-archive': ReplayArchive,
    'push-inbox': PushInbox
  };

  exports.run = function(router, request, cb) {
    var op, opClass, opName;
    opName = request.operation;
    if (!opName) {
      return typeof cb === "function" ? cb() : void 0;
    }
    opClass = OPERATIONS[opName];
    if (!opClass) {
      logger.warn("Unimplemented operation " + opName + ": " + (inspect(request)));
      return typeof cb === "function" ? cb(new errors.FeatureNotImplemented("Unimplemented operation " + opName)) : void 0;
    }
    logger.debug("operations.run: " + (inspect(request)));
    op = new opClass(router, request);
    return op.run(function(error, result) {
      if (error) {
        logger.warn("Operation " + opName + " failed: " + (error.stack || error));
        return typeof cb === "function" ? cb(error) : void 0;
      } else {
        logger.debug("Operation " + opName + " ran: " + (inspect(result)));
        return async.series([
          function(cb2) {
            var node, notification, notifications, _ref1, _results;
            if ((notification = typeof op.notification === "function" ? op.notification() : void 0)) {
              notification = notification.concat(generateSubscriptionsNotifications(notification));
              _ref1 = groupByNode(notification);
              _results = [];
              for (node in _ref1) {
                if (!__hasProp.call(_ref1, node)) continue;
                notifications = _ref1[node];
                notifications.node = node;
                _results.push(new Notify(router, notifications).run(function(err) {
                  if (err) {
                    logger.error("Error running notifications: " + (err.stack || err.message || err));
                  }
                  return cb2();
                }));
              }
              return _results;
            } else {
              return cb2();
            }
          }, function(cb2) {
            var notification;
            if ((notification = typeof op.moderatorNotification === "function" ? op.moderatorNotification() : void 0)) {
              return new ModeratorNotify(router, notification).run(function(err) {
                if (err) {
                  logger.error("Error running notifications: " + (err.stack || err.message || err));
                }
                return cb2();
              });
            } else {
              return cb2();
            }
          }, function(cb2) {
            var listener, node, req, user, _i, _len, _ref1, _ref2, _results;
            if (op.newModerators && op.newModerators.length > 0) {
              _ref1 = op.newModerators;
              _results = [];
              for (_i = 0, _len = _ref1.length; _i < _len; _i++) {
                _ref2 = _ref1[_i], user = _ref2.user, node = _ref2.node, listener = _ref2.listener;
                req = {
                  operation: 'new-moderator-notification',
                  node: node,
                  actor: user,
                  listener: listener
                };
                _results.push(new NewModeratorNotify(router, req).run(function(err) {
                  if (err) {
                    logger.error("Error running new moderator notification: " + (err.stack || err.message || err));
                  }
                  return cb2();
                }));
              }
              return _results;
            } else {
              return cb2();
            }
          }, function(cb2) {
            if (op.newNodes != null) {
              return async.forEach(op.newNodes, function(node, cb3) {
                return router.syncNode(node, cb3);
              }, cb2);
            } else {
              return cb2();
            }
          }
        ], function() {
          try {
            return typeof cb === "function" ? cb(null, result) : void 0;
          } catch (e) {
            return logger.error(e.stack || e);
          }
        });
      }
    });
  };

  groupByNode = function(updates) {
    var result, update, _i, _len;
    result = {};
    for (_i = 0, _len = updates.length; _i < _len; _i++) {
      update = updates[_i];
      if (!result.hasOwnProperty(update.node)) {
        result[update.node] = [];
      }
      result[update.node].push(update);
    }
    return result;
  };

  generateSubscriptionsNotifications = function(updates) {
    var itemIdsSeen;
    itemIdsSeen = {};
    return updates.filter(function(update) {
      return (update.type === 'subscription' && update.subscription === 'subscribed') || update.type === 'affiliation';
    }).map(function(update) {
      var followee, _ref1;
      followee = (_ref1 = update.node.match(NODE_OWNER_TYPE_REGEXP)) != null ? _ref1[1] : void 0;
      return {
        type: 'items',
        node: "/user/" + update.user + "/subscriptions",
        items: [
          {
            id: followee
          }
        ]
      };
    }).filter(function(update) {
      var itemId;
      itemId = update.items[0].id;
      if (itemIdsSeen[itemId] != null) {
        return false;
      } else {
        itemIdsSeen[itemId] = true;
        return true;
      }
    });
  };

}).call(this);
