// Generated by CoffeeScript 1.3.3
(function() {
  var async, config, inspect, path, version,
    __slice = [].slice;

  path = require('path');

  async = require('async');

  inspect = require('util').inspect;

  config = require('jsconfig');

  version = '0.3.1'

  config.defaults(path.join(__dirname, "config.js"));

  process.title = "buddycloud-server " + version;

  config.set('ignore unknown', true);

  config.set('env', {
    HOST: 'xmpp.host',
    PORT: ['xmpp.port', parseInt]
  });

  config.cli({
    host: ['xmpp.host', ['b', "xmpp server listen address", 'host']],
    port: ['xmpp.port', ['p', "xmpp server listen port", 'number']],
    config: ['c', "load config file", 'path'],
    debug: [false, "enable debug mode"],
    nobuild: [false, "[INTERNAL] disable build"],
    stdout: ['logging.stdout', [false, "Log to stdout"]],
    version: [false, "Display version"]
  });

  config.load("/etc/buddycloud-server/config.js", function(args, opts) {
    var errors, logger, logger_, model, pubsubBackend, pubsubServer, router, xmppConn, _ref;
    if (opts.version) {
      console.log(version);
      process.exit(0);
    }
    if ((_ref = opts.config) != null ? _ref.length : void 0) {
      if (opts.config[0] !== '/') {
        opts.config = path.join(process.cwd(), opts.config);
      }
      config.merge(opts.config);
    }
    if (opts.stdout) {
      config.logging.stdout = true;
    }
    logger_ = require('./logger');
    logger_.setConfig(config.logging);
    logger = logger_.makeLogger('main');
    if (opts.debug) {
      process.on('uncaughtException', function(err) {
        return logger.error("uncaughtException: " + (err.stack || err.message || err.toString()));
      });
    }
    errors = require('./errors');
    model = require('./local/model_postgres');
    model.start(config.modelConfig);
    router = new (require('./router').Router)(model, config.checkCreateNode, config.autosubscribeNewUsers);
    xmppConn = new (require('./xmpp/connection').Connection)(config.xmpp);
    pubsubServer = new (require('./xmpp/pubsub_server').PubsubServer)(xmppConn);
    pubsubBackend = new (require('./xmpp/backend_pubsub').PubsubBackend)(xmppConn);
    router.addBackend(pubsubBackend);
    pubsubServer.on('request', function(request) {
      logger.trace("request: " + (inspect(request)));
      if (request.operation === 'get-version') {
        return request.callback(null, {
          name: "buddycloud-server",
          version: version,
          os: process.platform
        });
      } else if (request.sender !== request.actor) {
        return pubsubBackend.authorizeFor(request.sender, request.actor, function(err, valid) {
          if (err) {
            return request.callback(err);
          } else if (!valid) {
            return request.callback(new errors.BadRequest('Requesting service not authorized for actor'));
          } else {
            return router.run(request, function() {
              var args;
              args = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
              return request.callback.apply(request, args);
            });
          }
        });
      } else {
        return router.run(request, function() {
          var args;
          args = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
          return request.callback.apply(request, args);
        });
      }
    });
    pubsubBackend.on('notificationPush', function(opts) {
      logger.trace("notificationPush: " + (inspect(opts)));
      opts.operation = 'push-inbox';
      return router.run(opts, function() {});
    });
    pubsubBackend.on('syncNeeded', function(server) {
      return router.syncServer(server, function() {});
    });
    pubsubBackend.on('authorizationPrompt', function(opts) {
      return pubsubBackend.authorizeFor(opts.sender, opts.nodeOwner, function(err, valid) {
        if (valid) {
          opts.type = 'authorizationPrompt';
          return router.notify(opts);
        }
      });
    });
    pubsubBackend.on('authorizationConfirm', function(opts) {
      opts.operation = 'confirm-subscriber-authorization';
      return router.run(opts, function() {});
    });
    xmppConn.on('userOffline', function(user) {
      return router.onUserOffline(user);
    });
    return xmppConn.on('online', function() {
      var saidHello;
      logger.info("XMPP connection established");
      process.title = "buddycloud-server " + version + ": " + xmppConn.jid;
      saidHello = false;
      model.forListeners(function(listener) {
        if (!saidHello) {
          logger.info("server successfully started");
          saidHello = true;
        }
        return xmppConn.probePresence(listener);
      });
      return router.setupSync(Math.ceil((config.modelConfig.poolSize || 2) / 2));
    });
  });

}).call(this);
