// Generated by CoffeeScript 1.3.3
(function() {
  var CACHE_TIMEOUT, RSM, RemoteRouter, async, errors, getNodeUser, inspect, logger, nodeTypes, rsmWalk, sync, _ref,
    __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };

  logger = require('./logger').makeLogger('router');

  inspect = require('util').inspect;

  errors = require('./errors');

  sync = require('./sync');

  async = require('async');

  rsmWalk = require('./rsm_walk');

  RSM = require('./xmpp/rsm').RSM;

  _ref = require('./util'), getNodeUser = _ref.getNodeUser, nodeTypes = _ref.nodeTypes;

  CACHE_TIMEOUT = 60 * 1000;

  RemoteRouter = (function() {

    function RemoteRouter(router) {
      this.router = router;
      this.detectAnonymousUser = __bind(this.detectAnonymousUser, this);

      this.authorizeFor = __bind(this.authorizeFor, this);

      this.isUserOnline = __bind(this.isUserOnline, this);

      this.addBackend = __bind(this.addBackend, this);

      this.backends = [];
    }

    RemoteRouter.prototype.addBackend = function(backend) {
      return this.backends.push(backend);
    };

    RemoteRouter.prototype.getMyJids = function() {
      var jids;
      jids = [];
      this.backends.map(function(backend) {
        if (backend.getMyJids != null) {
          return jids.push.apply(jids, backend.getMyJids());
        }
      });
      return jids;
    };

    RemoteRouter.prototype.isUserOnline = function(user) {
      return this.backends.some(function(backend) {
        return backend.isUserOnline(user);
      });
    };

    RemoteRouter.prototype.run = function(opts, cb) {
      var backends, tryBackend,
        _this = this;
      backends = (function(func, args, ctor) {
        ctor.prototype = func.prototype;
        var child = new ctor, result = func.apply(child, args), t = typeof result;
        return t == "object" || t == "function" ? result || child : child;
      })(Array, this.backends, function(){});
      tryBackend = function() {
        var backend;
        backend = backends.shift();
        return backend.run(_this.router, opts, function(err, results) {
          if (err && backends.length > 0) {
            return tryBackend();
          } else {
            return cb(err, results);
          }
        });
      };
      return tryBackend();
    };

    RemoteRouter.prototype.notify = function(notification) {
      var backend, _i, _len, _ref1, _results;
      _ref1 = this.backends;
      _results = [];
      for (_i = 0, _len = _ref1.length; _i < _len; _i++) {
        backend = _ref1[_i];
        _results.push(backend.notify(notification));
      }
      return _results;
    };

    RemoteRouter.prototype.authorizeFor = function(sender, actor, cb) {
      var backends, tryBackend,
        _this = this;
      backends = (function(func, args, ctor) {
        ctor.prototype = func.prototype;
        var child = new ctor, result = func.apply(child, args), t = typeof result;
        return t == "object" || t == "function" ? result || child : child;
      })(Array, this.backends, function(){});
      tryBackend = function() {
        var backend;
        backend = backends.shift();
        return backend.authorizeFor(sender, actor, function(err, valid) {
          if ((err || !valid) && backends.length > 0) {
            return tryBackend();
          } else {
            return cb(err, valid);
          }
        });
      };
      return tryBackend();
    };

    RemoteRouter.prototype.detectAnonymousUser = function(user, cb) {
      var backends, tryBackend,
        _this = this;
      backends = (function(func, args, ctor) {
        ctor.prototype = func.prototype;
        var child = new ctor, result = func.apply(child, args), t = typeof result;
        return t == "object" || t == "function" ? result || child : child;
      })(Array, this.backends, function(){});
      tryBackend = function() {
        var backend;
        backend = backends.shift();
        return backend.detectAnonymousUser(user, function(err, isAnonymous) {
          if (err && backends.length > 0) {
            return tryBackend();
          } else {
            return cb(err, isAnonymous);
          }
        });
      };
      return tryBackend();
    };

    return RemoteRouter;

  })();

  exports.Router = (function() {

    function Router(model, checkCreateNode, autosubscribeNewUsers) {
      this.model = model;
      this.autosubscribeNewUsers = autosubscribeNewUsers;
      this.remote = new RemoteRouter(this);
      this.operations = require('./local/operations');
      this.operations.setModel(model);
      this.operations.checkCreateNode = checkCreateNode;
      this.addBackend = this.remote.addBackend;
      this.isUserOnline = this.remote.isUserOnline;
      this.authorizeFor = this.remote.authorizeFor;
      this.anonymousUsers = {};
      this.perNodeQueue = {};
    }

    Router.prototype.detectUserType = function(user, cb) {
      var _this = this;
      if (user.indexOf("@") >= 0) {
        if (this.anonymousUsers.hasOwnProperty(user) && this.anonymousUsers[user]) {
          return cb(null, true);
        } else {
          return this.remote.detectAnonymousUser(user, function(err, isAnonymous) {
            if (err && user.indexOf('@anon') >= 0) {
              return cb(null, 'anonymous');
            } else if (isAnonymous) {
              return cb(null, 'anonymous');
            } else {
              return cb(null, 'user');
            }
          });
        }
      } else {
        return cb(null, 'server');
      }
    };

    Router.prototype.isLocallySubscribed = function(node, cb) {
      return this.model.nodeExists(node, cb);
    };

    Router.prototype.runLocally = function(opts, cb) {
      var _this = this;
      if (opts.node && this.perNodeQueue.hasOwnProperty(opts.node)) {
        return this.perNodeQueue[opts.node].push(function() {
          return _this.runLocally(opts, cb);
        });
      } else {
        return this.operations.run(this, opts, cb);
      }
    };

    Router.prototype.runRemotely = function(opts, cb) {
      return this.remote.run(opts, cb);
    };

    Router.prototype.run = function(opts, cb) {
      var _this = this;
      logger.trace("Router.run " + (inspect(opts)));
      return this.runCheckAnonymous(opts, function(err) {
        if (err) {
          return typeof cb === "function" ? cb(err) : void 0;
        }
        logger.info("run " + opts.operation + " " + opts.node + " by " + opts.actor + "(" + opts.actorType + ")/" + opts.sender);
        if (opts.node == null) {
          return _this.runLocally(opts, function(err, results) {
            var req, type, userid, _i, _len, _ref1, _results;
            if (typeof cb === "function") {
              cb(err, results);
            }
            if (!(err != null) && opts.operation === 'register-user' && (_this.autosubscribeNewUsers != null)) {
              _ref1 = _this.autosubscribeNewUsers;
              _results = [];
              for (_i = 0, _len = _ref1.length; _i < _len; _i++) {
                userid = _ref1[_i];
                _results.push((function() {
                  var _j, _len1, _results1;
                  _results1 = [];
                  for (_j = 0, _len1 = nodeTypes.length; _j < _len1; _j++) {
                    type = nodeTypes[_j];
                    req = Object.create(opts);
                    req.operation = 'subscribe-node';
                    req.node = "/user/" + userid + "/" + type;
                    req.writes = true;
                    _results1.push(this.run(req));
                  }
                  return _results1;
                }).call(_this));
              }
              return _results;
            }
          });
        } else if (opts.writes) {
          return _this.runRemotely(opts, function(err, results) {
            if (err && err.constructor === errors.SeeLocal) {
              return _this.runLocally(opts, cb);
            } else {
              return typeof cb === "function" ? cb(err, results) : void 0;
            }
          });
        } else {
          return _this.isLocallySubscribed(opts.node, function(err, locallySubscribed) {
            if (locallySubscribed) {
              return _this.runLocally(opts, cb);
            } else {
              return _this.runRemotely(opts, function(err, results) {
                if ((err != null ? err.constructor : void 0) === errors.SeeLocal) {
                  return typeof cb === "function" ? cb(new errors.NotFound("Node does not exist here")) : void 0;
                } else {
                  return typeof cb === "function" ? cb(err, results) : void 0;
                }
              });
            }
          });
        }
      });
    };

    Router.prototype.runCheckAnonymous = function(opts, cb) {
      var _this = this;
      if (opts.actorType) {
        if (!opts.writes) {
          return cb();
        } else {
          if (opts.actorType === 'anonymous') {
            if (opts.operation === 'subscribe-node' || opts.operation === 'unsubscribe-node') {
              if (this.isUserOnline(opts.actor)) {
                this.anonymousUsers[opts.actor] = true;
                return cb();
              } else {
                return cb(new errors.Forbidden("Send presence to be able to temporarily subscribe."));
              }
            } else {
              return cb(new errors.NotAllowed("You are anonymous. You are legion."));
            }
          } else {
            return cb();
          }
        }
      } else {
        return this.detectUserType(opts.actor, function(err, type) {
          opts.actorType = type || 'user';
          return _this.runCheckAnonymous(opts, cb);
        });
      }
    };

    Router.prototype.pushData = function(opts, cb) {
      opts.operation = 'push-inbox';
      return this.runLocally(opts, cb);
    };

    Router.prototype.notify = function(notification) {
      return this.remote.notify(notification);
    };

    Router.prototype.setupSync = function(jobs) {
      var _this = this;
      sync.setup(jobs);
      return this.model.getAllNodes(function(err, nodes) {
        var node, _i, _len, _results;
        if (err) {
          logger.error("getAllNodes " + (err.stack || err));
          return;
        }
        _results = [];
        for (_i = 0, _len = nodes.length; _i < _len; _i++) {
          node = nodes[_i];
          _results.push(_this.syncNode(node, function(err) {
            if (err) {
              return logger.error("syncNode " + node + " error: " + err);
            }
          }));
        }
        return _results;
      });
    };

    Router.prototype.syncNode = function(node, cb) {
      var _this = this;
      if (!this.perNodeQueue.hasOwnProperty(node)) {
        this.perNodeQueue[node] = [];
      }
      return sync.syncNode(this, this.model, node, function(err) {
        var blocked;
        blocked = _this.perNodeQueue[node] || [];
        delete _this.perNodeQueue[node];
        blocked.forEach(function(cb1) {
          return cb1();
        });
        return cb(err);
      });
    };

    Router.prototype.syncServer = function(server, cb) {
      var opts,
        _this = this;
      opts = {
        operation: 'retrieve-user-subscriptions',
        jid: server
      };
      opts.rsm = new RSM();
      return rsmWalk(function(nextOffset, cb2) {
        opts.rsm.after = nextOffset;
        return _this.runRemotely(opts, function(err, results) {
          if (err) {
            return cb2(err);
          }
          return async.forEach(results, function(subscription, cb3) {
            var user;
            if (!subscription.node) {
              return cb3();
            }
            user = getNodeUser(subscription.node);
            if (!user) {
              return cb3();
            }
            return _this.authorizeFor(server, user, function(err, valid) {
              if (err || !valid) {
                logger.warn((err && err.stack) || err || ("Cannot sync " + subscription.node + " from unauthorized server " + server));
                return cb3();
              } else {
                return _this.syncNode(subscription.node, function(err) {
                  return cb3();
                });
              }
            });
          }, function(err) {
            var _ref1;
            return cb2(err, results != null ? (_ref1 = results.rsm) != null ? _ref1.last : void 0 : void 0);
          });
        });
      }, cb);
    };

    Router.prototype.onUserOffline = function(user) {
      var req;
      if (this.anonymousUsers.hasOwnProperty(user) && this.anonymousUsers[user]) {
        delete this.anonymousUsers[user];
        req = {
          operation: 'remove-user',
          actor: user,
          sender: user
        };
        return runLocally(req, function() {});
      }
    };

    return Router;

  })();

}).call(this);
