// Generated by CoffeeScript 1.3.3
(function() {
  var BuddycloudDiscovery, EventEmitter, NS, Notifications, RequestCache, async, errors, forms, getNodeUser, getUserDomain, logger, pubsubClient, xmpp, _ref,
    __hasProp = {}.hasOwnProperty,
    __extends = function(child, parent) { for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; } function ctor() { this.constructor = child; } ctor.prototype = parent.prototype; child.prototype = new ctor(); child.__super__ = parent.prototype; return child; },
    __slice = [].slice,
    __bind = function(fn, me){ return function(){ return fn.apply(me, arguments); }; };

  logger = require('../logger').makeLogger('xmpp/backend_pubsub');

  EventEmitter = require('events').EventEmitter;

  async = require('async');

  xmpp = require('node-xmpp');

  Notifications = require('./pubsub_notifications');

  pubsubClient = require('./pubsub_client');

  errors = require('../errors');

  NS = require('./ns');

  forms = require('./forms');

  _ref = require('../util'), getNodeUser = _ref.getNodeUser, getUserDomain = _ref.getUserDomain;

  exports.PubsubBackend = (function(_super) {

    __extends(PubsubBackend, _super);

    function PubsubBackend(conn) {
      var _this = this;
      this.conn = conn;
      this.conn.on('message', function() {
        var args;
        args = 1 <= arguments.length ? __slice.call(arguments, 0) : [];
        return _this.onMessage_.apply(_this, args);
      });
      this.disco = new BuddycloudDiscovery(this.conn);
      this.authorizeFor = this.disco.authorizeFor;
      this.detectAnonymousUser = this.disco.detectAnonymousUser;
    }

    PubsubBackend.prototype.getMyJids = function() {
      return [this.conn.jid];
    };

    PubsubBackend.prototype.isUserOnline = function(user) {
      return this.conn.getOnlineResources(user).length > 0;
    };

    PubsubBackend.prototype.run = function(router, opts, cb) {
      var req, reqClass, user,
        _this = this;
      if (opts.jid != null) {
        reqClass = pubsubClient.byOperation(opts.operation);
        if (!reqClass) {
          return cb(new errors.FeatureNotImplemented("Operation " + opts.operation + " not implemented for remote pubsub"));
        }
        return req = new reqClass(this.conn, opts, function(err, result) {
          var localPushData;
          if (err) {
            return cb(err);
          } else {
            if ((localPushData = typeof req.localPushData === "function" ? req.localPushData() : void 0)) {
              return router.pushData(localPushData, function(err) {
                return cb(err, result);
              });
            } else {
              return cb(null, result);
            }
          }
        });
      } else {
        user = getNodeUser(opts.node);
        if (!user) {
          return cb(new errors.NotFound("Unrecognized node form"));
        }
        logger.debug({
          'PubsubBackend.run': opts,
          user: user
        });
        return this.disco.findService(user, function(err, service) {
          var opts2;
          if (err) {
            return cb(err);
          }
          if (_this.getMyJids().indexOf(service) >= 0) {
            return cb(new errors.SeeLocal());
          } else {
            opts2 = Object.create(opts);
            opts2.jid = service;
            return _this.run(router, opts2, cb);
          }
        });
      }
    };

    PubsubBackend.prototype.notify = function(opts) {
      var copyProps, listener, notification, onlineJid, opts1, opts2, pivot, _i, _len, _ref1, _results;
      if (opts.length < 1) {
        return;
      }
      notification = Notifications.make(opts);
      listener = opts.listener;
      try {
        if (!opts.replay && listener && listener.indexOf("@") >= 0) {
          _ref1 = this.conn.getOnlineResources(listener);
          _results = [];
          for (_i = 0, _len = _ref1.length; _i < _len; _i++) {
            onlineJid = _ref1[_i];
            logger.info("notifying client " + onlineJid + " for " + opts.node);
            _results.push(this.conn.send(notification.toStanza(this.conn.jid, onlineJid)));
          }
          return _results;
        } else if (listener) {
          logger.info("notifying " + listener + " for " + opts.node);
          return this.conn.send(notification.toStanza(this.conn.jid, listener));
        }
      } catch (e) {
        if (e.constructor === errors.MaxStanzaSizeExceeded && opts.length > 1) {
          copyProps = function(opts_) {
            return ['type', 'node', 'user', 'listener', 'replay'].forEach(function(prop) {
              return opts_[prop] = opts[prop];
            });
          };
          pivot = Math.floor(opts.length / 2);
          opts1 = opts.slice(0, pivot);
          copyProps(opts1);
          opts2 = opts.slice(pivot, opts.length);
          copyProps(opts2);
          logger.warn("MaxStanzaSizeExceeded: split notification from " + opts.length + " into " + opts1.length + "+" + opts2.length);
          this.notify(opts1);
          return this.notify(opts2);
        } else {
          throw e;
        }
      }
    };

    PubsubBackend.prototype.onMessage_ = function(message) {
      var actor, allow, child, form, node, nodeOwner, sender, updates, user, _i, _len, _ref1,
        _this = this;
      sender = new xmpp.JID(message.attrs.from).bare().toString();
      updates = [];
      _ref1 = message.children;
      for (_i = 0, _len = _ref1.length; _i < _len; _i++) {
        child = _ref1[_i];
        if (!child.is) {
          continue;
        }
        if (child.is("event", NS.PUBSUB_EVENT)) {
          child.children.forEach(function(child) {
            var config, form, item, itemEl, items, node, xEl, _j, _len1, _ref2;
            if (!child.is) {
              return;
            }
            node = child.attrs.node;
            if (!node) {
              return;
            }
            if (child.is('items')) {
              items = [];
              _ref2 = child.getChildren('item');
              for (_j = 0, _len1 = _ref2.length; _j < _len1; _j++) {
                itemEl = _ref2[_j];
                item = {
                  el: itemEl.children.filter(function(itemEl) {
                    return itemEl.hasOwnProperty('children');
                  })[0]
                };
                if (itemEl.attrs.id) {
                  item.id = itemEl.attrs.id;
                }
                if (item.el) {
                  items.push(item);
                }
              }
              updates.push({
                type: 'items',
                node: node,
                items: items
              });
            }
            if (child.is('subscription')) {
              updates.push({
                type: 'subscription',
                node: node,
                user: child.attrs.jid,
                subscription: child.attrs.subscription
              });
            }
            if (child.is('affiliation')) {
              updates.push({
                type: 'affiliation',
                node: node,
                user: child.attrs.jid,
                affiliation: child.attrs.affiliation
              });
            }
            if (child.is('configuration')) {
              xEl = child.getChild('x', NS.DATA);
              form = xEl && forms.fromXml(xEl);
              config = form && forms.formToConfig(form);
              if (config) {
                return updates.push({
                  type: 'config',
                  node: node,
                  config: config
                });
              }
            }
          });
        }
        if (child.is("you-missed-something", NS.BUDDYCLOUD_V1)) {
          this.emit('syncNeeded', sender);
        }
        if (child.is("x", NS.DATA)) {
          form = forms.fromXml(child);
          if (form.type === 'form' && form.getFormType() === NS.PUBSUB_SUBSCRIBE_AUTHORIZATION) {
            node = form.get('pubsub#node');
            user = form.get('pubsub#subscriber_jid');
            nodeOwner = getNodeUser(node);
            if (nodeOwner) {
              this.emit('authorizationPrompt', {
                node: node,
                user: user,
                sender: sender,
                nodeOwner: nodeOwner
              });
            }
          } else if (form.type === 'submit' && form.getFormType() === NS.PUBSUB_SUBSCRIBE_AUTHORIZATION) {
            node = form.get('pubsub#node');
            user = form.get('pubsub#subscriber_jid');
            allow = form.get('pubsub#allow') === 'true';
            actor = form.get('buddycloud#actor') || sender;
            this.emit('authorizationConfirm', {
              node: node,
              user: user,
              allow: allow,
              actor: actor,
              sender: sender
            });
          }
        }
      }
      return async.filter(updates, function(update, cb) {
        user = getNodeUser(update.node);
        if (!user) {
          return cb(false);
        }
        return _this.authorizeFor(sender, user, function(err, valid) {
          return cb(!err && valid);
        });
      }, function(updates) {
        if ((updates != null) && updates.length > 0) {
          updates.sender = sender;
          updates.actor = sender;
          return _this.emit('notificationPush', updates);
        }
      });
    };

    return PubsubBackend;

  })(EventEmitter);

  BuddycloudDiscovery = (function() {

    function BuddycloudDiscovery(conn) {
      var _this = this;
      this.conn = conn;
      this.detectAnonymousUser = __bind(this.detectAnonymousUser, this);

      this.findService = __bind(this.findService, this);

      this.authorizeFor = __bind(this.authorizeFor, this);

      this.infoCache = new RequestCache(function(id, cb) {
        return new pubsubClient.DiscoverInfo(_this.conn, {
          jid: id
        }, cb);
      });
      this.itemsCache = new RequestCache(function(id, cb) {
        logger.debug("discover items of " + id);
        return new pubsubClient.DiscoverItems(_this.conn, {
          jid: id
        }, cb);
      });
    }

    BuddycloudDiscovery.prototype.authorizeFor = function(sender, actor, cb) {
      return this.itemsCache.get(getUserDomain(actor), function(err, items) {
        var valid;
        if (err) {
          return cb(err);
        }
        valid = items != null ? items.some(function(item) {
          return item.jid === sender;
        }) : void 0;
        logger.debug("authorizing " + sender + " for " + actor + ": " + valid);
        return cb(null, valid);
      });
    };

    BuddycloudDiscovery.prototype.findService = function(user, cb) {
      var domain,
        _this = this;
      domain = getUserDomain(user);
      return this.itemsCache.get(domain, function(err, items) {
        var done, pending, resultSent;
        if (err) {
          return cb(err);
        }
        pending = 1;
        resultSent = false;
        done = function() {
          pending--;
          if (pending < 1 && !resultSent) {
            resultSent = true;
            return cb(new errors.NotFound("No pubsub channel service discovered for " + user));
          }
        };
        items.forEach(function(item) {
          _this.infoCache.get(item.jid, function(err, result) {
            var identity, _i, _len, _ref1;
            _ref1 = (result != null ? result.identities : void 0) || [];
            for (_i = 0, _len = _ref1.length; _i < _len; _i++) {
              identity = _ref1[_i];
              if (identity.category === "pubsub" && identity.type === "channels" && !resultSent) {
                resultSent = true;
                logger.debug("found service for " + user + ": " + item.jid);
                cb(null, item.jid);
              }
            }
            return done();
          });
          return pending++;
        });
        return done();
      });
    };

    BuddycloudDiscovery.prototype.detectAnonymousUser = function(user, cb) {
      return this.infoCache.get(user, function(err, result) {
        var isAnonymous, _ref1;
        isAnonymous = false;
        if (typeof results !== "undefined" && results !== null) {
          if ((_ref1 = results.identities) != null) {
            _ref1.forEach(function(identity) {
              return isAnonymous || (isAnonymous = identity.category === "account" && identity.type === "anonymous");
            });
          }
        }
        return cb(err, isAnonymous);
      });
    };

    return BuddycloudDiscovery;

  })();

  RequestCache = (function() {

    RequestCache.prototype.cacheTimeout = 30 * 1000;

    function RequestCache(getter) {
      this.getter = getter;
      this.entries = {};
    }

    RequestCache.prototype.get = function(id, cb) {
      var _this = this;
      if (!this.entries.hasOwnProperty(id)) {
        logger.trace("Cache miss for " + id);
        this.entries[id] = {
          queued: [cb]
        };
        return this.getter(id, function(err, results) {
          var cb1, queued, _i, _len, _results;
          queued = _this.entries[id].queued;
          _this.entries[id] = {
            err: err,
            results: results
          };
          setTimeout(function() {
            return delete _this.entries[id];
          }, _this.cacheTimeout);
          _results = [];
          for (_i = 0, _len = queued.length; _i < _len; _i++) {
            cb1 = queued[_i];
            _results.push(cb1(err, results));
          }
          return _results;
        });
      } else if (this.entries[id].queued != null) {
        logger.trace("Queued for " + id);
        return this.entries[id].queued.push(cb);
      } else {
        logger.trace("Cache hit for " + id);
        return process.nextTick(function() {
          return cb(_this.entries[id].err, _this.entries[id].results);
        });
      }
    };

    return RequestCache;

  })();

}).call(this);
