// Generated by CoffeeScript 1.3.3

/*
# Encapsulate XMPP Component connection
#
# * Provide <iq/> RPC interface
# * Track presence
*/


(function() {
  var EventEmitter, IQ_TIMEOUT, MAX_STANZA_SIZE, MISSED_INTERVAL, NS, errors, logger, xmpp,
    __hasProp = {}.hasOwnProperty,
    __extends = function(child, parent) { for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; } function ctor() { this.constructor = child; } ctor.prototype = parent.prototype; child.prototype = new ctor(); child.__super__ = parent.prototype; return child; };

  logger = require('../logger').makeLogger('xmpp/connection');

  xmpp = require("node-xmpp");

  EventEmitter = require('events').EventEmitter;

  errors = require("../errors");

  NS = require('./ns');

  IQ_TIMEOUT = 10000;

  MAX_STANZA_SIZE = 65000;

  MISSED_INTERVAL = 120 * 1000;

  exports.Connection = (function(_super) {

    __extends(Connection, _super);

    function Connection(config) {
      var _this = this;
      this.lastIqId = 0;
      this.iqCallbacks = {};
      this.missedServerTimeouts = {};
      this.onlineResources = {};
      this.jid = config.jid;
      config.reconnect = true;
      this.conn = new xmpp.Component(config);
      this.conn.on("error", function(e) {
        return logger.error(e);
      });
      this.conn.on("online", function() {
        return _this.emit("online");
      });
      this.conn.on("stanza", function(stanza) {
        var cb, from;
        logger.trace("<< " + (stanza.toString()));
        from = stanza.attrs.from;
        switch (stanza.name) {
          case "iq":
            switch (stanza.attrs.type) {
              case "get":
              case "set":
                return _this._handleIq(stanza);
              case "result":
              case "error":
                _this.iqCallbacks.hasOwnProperty(stanza.attrs.id);
                cb = _this.iqCallbacks[stanza.attrs.id];
                delete _this.iqCallbacks[stanza.attrs.id];
                if (stanza.attrs.type === 'error') {
                  return cb && cb(new errors.StanzaError(stanza));
                } else {
                  return cb && cb(null, stanza);
                }
            }
            break;
          case "presence":
            return _this._handlePresence(stanza);
          case "message":
            if (stanza.attrs.type !== "error") {
              return _this._handleMessage(stanza);
            } else if (from.indexOf('@') < 0) {
              if (!_this.missedServerTimeouts.hasOwnProperty(from)) {
                return _this.missedServerTimeouts[from] = setTimeout(function() {
                  delete _this.missedServerTimeouts[from];
                  return _this.send(new xmpp.Element('message', {
                    type: 'headline',
                    from: _this.jid,
                    to: from
                  }).c('you-missed-something', {
                    xmlns: NS.BUDDYCLOUD_V1
                  }));
                }, MISSED_INTERVAL);
              }
            }
        }
      });
    }

    Connection.prototype.send = function(stanza) {
      var bytes, s;
      stanza = stanza.root();
      if (!stanza.attrs.from) {
        stanza.attrs.from = this.jid;
      }
      bytes = 0;
      stanza.root().write(function(s) {
        return bytes += Buffer.byteLength(s);
      });
      if (bytes > MAX_STANZA_SIZE) {
        logger.warn("Stanza with " + bytes + " bytes: " + (stanza.toString().substr(0, 1024)) + "...");
        throw new errors.MaxStanzaSizeExceeded(bytes);
      }
      s = stanza.toString();
      logger.trace(">> " + s);
      if (stanza.attrs.to === this.jid) {
        return this.conn.emit('stanza', stanza);
      } else {
        return this.conn.send(s);
      }
    };

    Connection.prototype.sendIq = function(iq, cb) {
      var id, timeout,
        _this = this;
      this.lastIqId += Math.ceil(Math.random() * 999);
      id = iq.attrs.id = "" + this.lastIqId;
      timeout = setTimeout(function() {
        delete _this.iqCallbacks[id];
        return cb(new Error('timeout'));
      }, IQ_TIMEOUT);
      this.iqCallbacks[id] = function(error, result) {
        clearTimeout(timeout);
        return cb(error, result);
      };
      return this.send(iq);
    };

    Connection.prototype._handleMessage = function(message) {
      return this.emit('message', message);
    };

    Connection.prototype.getOnlineResources = function(bareJid) {
      if (this.onlineResources.hasOwnProperty(bareJid)) {
        return this.onlineResources[bareJid].map(function(resource) {
          var jid;
          jid = new xmpp.JID(bareJid);
          jid.resource = resource;
          return jid.toString();
        });
      } else {
        return [];
      }
    };

    Connection.prototype.subscribePresence = function(jid) {
      if (!this.onlineResources.hasOwnProperty(jid)) {
        return this.send(new xmpp.Element("presence", {
          to: jid,
          type: "subscribe"
        }));
      }
    };

    Connection.prototype._handlePresence = function(presence) {
      var bareJid, jid, resource, rmUserResource,
        _this = this;
      jid = new xmpp.JID(presence.attrs.from);
      bareJid = jid.bare().toString();
      resource = jid.resource;
      rmUserResource = function() {
        if (_this.onlineResources[bareJid] != null) {
          _this.onlineResources[bareJid] = _this.onlineResources[bareJid].filter(function(r) {
            return r !== resource;
          });
          if (_this.onlineResources[bareJid].length < 1) {
            delete _this.onlineResources[bareJid];
            return _this.emit('userOffline', bareJid);
          }
        }
      };
      switch (presence.attrs.type) {
        case "subscribe":
          return this.send(new xmpp.Element("presence", {
            from: presence.attrs.to,
            to: presence.attrs.from,
            id: presence.attrs.id,
            type: "subscribed"
          }));
        case "unsubscribe":
          return this.send(new xmpp.Element("presence", {
            from: presence.attrs.to,
            to: presence.attrs.from,
            id: presence.attrs.id,
            type: "unsubscribed"
          }));
        case "probe":
          return this.send(new xmpp.Element("presence", {
            from: presence.attrs.to,
            to: presence.attrs.from,
            id: presence.attrs.id
          }).c("status").t("buddycloud-server"));
        case "subscribed":
          break;
        case "unsubscribed":
          break;
        case "error":
          if (!resource) {
            return delete this.onlineResources[bareJid];
          } else {
            return rmUserResource();
          }
          break;
        case "unavailable":
          return rmUserResource();
        default:
          if (!(bareJid in this.onlineResources)) {
            this.onlineResources[bareJid] = [];
            this.emit('userOnline', bareJid);
          }
          if (this.onlineResources[bareJid].indexOf(resource) < 0) {
            return this.onlineResources[bareJid].push(resource);
          }
      }
    };

    Connection.prototype._handleIq = function(stanza) {
      var replied, replying,
        _this = this;
      replied = false;
      replying = function() {
        if (replied) {
          throw 'Sending additional iq reply';
        }
      };
      stanza.reply = function(child) {
        var reply;
        replying();
        reply = new xmpp.Element("iq", {
          from: stanza.attrs.to,
          to: stanza.attrs.from,
          id: stanza.attrs.id || "",
          type: "result"
        });
        if ((child != null ? child.children : void 0) != null) {
          reply.cnode(child.root());
        }
        _this.send(reply);
        return replied = true;
      };
      stanza.replyError = function(err) {
        var reply;
        replying();
        reply = new xmpp.Element("iq", {
          from: stanza.attrs.to,
          to: stanza.attrs.from,
          id: stanza.attrs.id || "",
          type: "error"
        });
        if (err.xmppElement) {
          reply.cnode(err.xmppElement());
        } else {
          reply.c("error", {
            type: "cancel"
          }).c("text").t('' + err.message);
        }
        return _this.send(reply);
      };
      return this.emit('iqRequest', stanza);
    };

    Connection.prototype.probePresence = function(user) {
      var sendPresence,
        _this = this;
      sendPresence = function(type) {
        return _this.conn.send(new xmpp.Element('presence', {
          type: type,
          from: _this.conn.jid,
          to: user
        }));
      };
      return sendPresence('probe');
    };

    return Connection;

  })(EventEmitter);

}).call(this);
