// Generated by CoffeeScript 1.3.3
(function() {
  var NS, logger, xmpp;

  logger = require('../logger').makeLogger('xmpp/forms');

  xmpp = require('node-xmpp');

  NS = require('./ns');

  exports.Field = (function() {

    function Field(_var, type, label, value) {
      this["var"] = _var != null ? _var : '';
      this.type = type != null ? type : 'text-single';
      this.label = label;
      this.values = [];
      if (value) {
        this.values.push(value);
      }
    }

    Field.prototype.toXml = function() {
      var addValue, fieldAttrs, fieldEl, _ref, _ref1, _ref2;
      fieldAttrs = {};
      if ((_ref = fieldAttrs["var"]) == null) {
        fieldAttrs["var"] = this["var"];
      }
      if ((_ref1 = fieldAttrs.label) == null) {
        fieldAttrs.label = this.label;
      }
      if ((_ref2 = fieldAttrs.type) == null) {
        fieldAttrs.type = this.type;
      }
      fieldEl = new xmpp.Element('field', fieldAttrs);
      addValue = function(value) {
        return fieldEl.c('value').t(value);
      };
      if (!/-multi$/.test(this.type) && (this.values[0] != null)) {
        addValue(this.values[0]);
      } else {
        this.values.forEach(addValue);
      }
      return fieldEl;
    };

    return Field;

  })();

  exports.Form = (function() {

    function Form(type, formType) {
      this.type = type != null ? type : 'result';
      this.fields = [];
      if (formType) {
        this.fields.push(new exports.Field('FORM_TYPE', 'hidden', void 0, formType));
      }
    }

    Form.prototype.getFormType = function() {
      var field, _i, _len, _ref;
      _ref = this.fields;
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        field = _ref[_i];
        if (field["var"] === 'FORM_TYPE') {
          return field.values[0];
        }
      }
      return null;
    };

    Form.prototype.get = function(fieldVar) {
      var field, _i, _len, _ref;
      _ref = this.fields;
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        field = _ref[_i];
        if (field["var"] === fieldVar) {
          return field.values[0];
        }
      }
      return null;
    };

    Form.prototype.addField = function(var_, type, label, value) {
      return this.fields.push(new exports.Field(var_, type, label, value));
    };

    Form.prototype.toXml = function() {
      var formEl;
      formEl = new xmpp.Element('x', {
        xmlns: NS.DATA,
        type: this.type
      });
      if (this.title != null) {
        formEl.c('title').t(this.title);
      }
      if (this.instructions != null) {
        formEl.c('instructions').t(this.instructions);
      }
      this.fields.forEach(function(field) {
        return formEl.cnode(field.toXml());
      });
      return formEl;
    };

    return Form;

  })();

  exports.fromXml = function(xEl) {
    var form;
    if (!xEl.is('x', NS.DATA)) {
      logger.warn("Importing non-form: " + (xEl.toString()));
    }
    form = new exports.Form(xEl.attrs.type);
    form.fields = xEl.getChildren("field").map(function(fieldEl) {
      var field;
      field = new exports.Field(fieldEl.attrs["var"], fieldEl.attrs.type, fieldEl.attrs.label);
      field.values = fieldEl.getChildren("value").map(function(valueEl) {
        return valueEl.getText();
      });
      return field;
    });
    return form;
  };

  exports.configToForm = function(config, type, formType) {
    var addField, form;
    form = new exports.Form(type, formType);
    addField = function(key, fvar, label) {
      if (config[key]) {
        return form.fields.push(new exports.Field(fvar, 'text-single', label, config[key]));
      }
    };
    addField('title', 'pubsub#title', 'A short name for the node');
    addField('description', 'pubsub#description', 'A description of the node');
    addField('accessModel', 'pubsub#access_model', 'Who may subscribe and retrieve items');
    addField('publishModel', 'pubsub#publish_model', 'Who may publish items');
    addField('defaultAffiliation', 'buddycloud#default_affiliation', 'What role do new subscribers have?');
    addField('creationDate', 'pubsub#creation_date', 'Creation date');
    addField('channelType', 'buddycloud#channel_type', 'Type of channel');
    return form;
  };

  exports.formToConfig = function(form) {
    var config, _ref, _ref1, _ref2, _ref3, _ref4, _ref5, _ref6;
    config = null;
    if ((form.getFormType() === NS.PUBSUB_NODE_CONFIG || form.getFormType() === NS.PUBSUB_META_DATA) && (form.type === 'submit' || form.type === 'result')) {
      config = {};
      if ((_ref = config.title) == null) {
        config.title = form.get('pubsub#title');
      }
      if ((_ref1 = config.description) == null) {
        config.description = form.get('pubsub#description');
      }
      if ((_ref2 = config.accessModel) == null) {
        config.accessModel = form.get('pubsub#access_model');
      }
      if ((_ref3 = config.publishModel) == null) {
        config.publishModel = form.get('pubsub#publish_model');
      }
      if ((_ref4 = config.defaultAffiliation) == null) {
        config.defaultAffiliation = form.get('buddycloud#default_affiliation');
      }
      if ((_ref5 = config.creationDate) == null) {
        config.creationDate = form.get('pubsub#creation_date');
      }
      if ((_ref6 = config.channelType) == null) {
        config.channelType = form.get('buddycloud#channel_type');
      }
    }
    return config;
  };

}).call(this);
